# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import os
import pathlib
import tempfile
from abc import ABC, abstractmethod
from typing import Any, Optional, Union

import pandas as pd

from mlrun.artifacts import Artifact, DatasetArtifact
from mlrun.datastore import DataItem
from mlrun.errors import MLRunInvalidArgumentError

from ..utils import ArtifactType, SupportedFormat
from .default_packager import DefaultPackager


class _Formatter(ABC):
    """
    An abstract class for a pandas formatter - supporting saving and loading dataframes to and from specific file type.
    """

    @classmethod
    @abstractmethod
    def to(
        cls, obj: pd.DataFrame, file_path: str, flatten: bool = True, **to_kwargs
    ) -> dict:
        """
        Save the given dataframe to the file path given.

        :param obj:       The dataframe to save.
        :param file_path: The file to save to.
        :param flatten:   Whether to flatten the dataframe before saving. For some formats it is mandatory to enable
                          flattening, otherwise saving and loading the dataframe will cause unexpected behavior
                          especially in case it is multi-level or multi-index. Default to True.
        :param to_kwargs: Additional keyword arguments to pass to the relevant `to_x` function.

        :return A dictionary of keyword arguments for reading the dataframe from file.
        """
        pass

    @classmethod
    @abstractmethod
    def read(
        cls, file_path: str, unflatten_kwargs: Optional[dict] = None, **read_kwargs
    ) -> pd.DataFrame:
        """
        Read the dataframe from the given file path.

        :param file_path:        The file to read the dataframe from.
        :param unflatten_kwargs: Unflatten keyword arguments for unflattening the read dataframe.
        :param read_kwargs:      Additional keyword arguments to pass to the relevant read function of pandas.

        :return: The loaded dataframe.
        """
        pass

    @staticmethod
    def _flatten_dataframe(dataframe: pd.DataFrame) -> tuple[pd.DataFrame, dict]:
        """
        Flatten the dataframe: moving all indexes to be columns at the start (from column 0) and lowering the columns
        levels to 1, renaming them from tuples. All columns and index info is stored so it can be unflatten later on.

        :param dataframe: The dataframe to flatten.

        :return: The flat dataframe.
        """
        # Save columns info:
        columns = list(dataframe.columns)
        if isinstance(dataframe.columns, pd.MultiIndex):
            columns = [list(column_tuple) for column_tuple in columns]
        columns_levels = list(dataframe.columns.names)

        # Save index info:
        index_levels = list(dataframe.index.names)

        # Turn multi-index columns into single columns:
        if len(columns_levels) > 1:
            # We turn the column tuple into a string to eliminate parsing issues during savings to text formats:
            dataframe.columns = pd.Index(
                "-".join(column_tuple) for column_tuple in columns
            )

        # Rename indexes in case they appear in the columns so it won't get overriden when the index reset:
        dataframe.index.set_names(
            names=[
                name
                if name is not None and name not in dataframe.columns
                else f"INDEX_{name}_{i}"
                for i, name in enumerate(dataframe.index.names)
            ],
            inplace=True,
        )

        # Reset the index, moving the current index to a column:
        dataframe.reset_index(inplace=True)

        return dataframe, {
            "columns": columns,
            "columns_levels": columns_levels,
            "index_levels": index_levels,
        }

    @staticmethod
    def _unflatten_dataframe(
        dataframe: pd.DataFrame,
        columns: list,
        columns_levels: list,
        index_levels: list,
    ) -> pd.DataFrame:
        """
        Unflatten the dataframe, moving the indexes from the columns and resuming the columns levels and names.

        :param dataframe:      The dataframe to unflatten.
        :param columns:        The original list of columns.
        :param columns_levels: The original columns levels names.
        :param index_levels:   The original index levels names.

        :return: The un-flatted dataframe.
        """
        # Move back index from columns:
        dataframe.set_index(
            keys=list(dataframe.columns[: len(index_levels)]), inplace=True
        )
        dataframe.index.set_names(names=index_levels, inplace=True)

        # Set the columns back in case they were multi-leveled:
        if len(columns_levels) > 1:
            dataframe.columns = pd.MultiIndex.from_tuples(
                tuples=columns, names=columns_levels
            )
        else:
            dataframe.columns.set_names(names=columns_levels, inplace=True)

        return dataframe


class _ParquetFormatter(_Formatter):
    """
    A static class for managing pandas parquet files.
    """

    @classmethod
    def to(
        cls, obj: pd.DataFrame, file_path: str, flatten: bool = True, **to_kwargs
    ) -> dict:
        """
        Save the given dataframe to the parquet file path given.

        :param obj:       The dataframe to save.
        :param file_path: The file to save to.
        :param flatten:   Ignored for parquet format.
        :param to_kwargs: Additional keyword arguments to pass to the `to_parquet` function.

        :return A dictionary of keyword arguments for reading the dataframe from file.
        """
        obj.to_parquet(path=file_path, **to_kwargs)
        return {}

    @classmethod
    def read(
        cls, file_path: str, unflatten_kwargs: Optional[dict] = None, **read_kwargs
    ) -> pd.DataFrame:
        """
        Read the dataframe from the given parquet file path.

        :param file_path:        The file to read the dataframe from.
        :param unflatten_kwargs: Ignored for parquet format.
        :param read_kwargs:      Additional keyword arguments to pass to the `read_parquet` function.

        :return: The loaded dataframe.
        """
        return pd.read_parquet(path=file_path, **read_kwargs)


class _CSVFormatter(_Formatter):
    """
    A static class for managing pandas csv files.
    """

    @classmethod
    def to(
        cls, obj: pd.DataFrame, file_path: str, flatten: bool = True, **to_kwargs
    ) -> dict:
        """
        Save the given dataframe to the csv file path given.

        :param obj:       The dataframe to save.
        :param file_path: The file to save to.
        :param flatten:   Whether to flatten the dataframe before saving. For some formats it is mandatory to enable
                          flattening, otherwise saving and loading the dataframe will cause unexpected behavior
                          especially in case it is multi-level or multi-index. Default to True.
        :param to_kwargs: Additional keyword arguments to pass to the `to_csv` function.

        :return A dictionary of keyword arguments for reading the dataframe from file.
        """
        # Flatten the dataframe (this format have problems saving multi-level dataframes):
        instructions = {}
        if flatten:
            obj, unflatten_kwargs = cls._flatten_dataframe(dataframe=obj)
            instructions["unflatten_kwargs"] = unflatten_kwargs

        # Write to csv:
        obj.to_csv(path_or_buf=file_path, **to_kwargs)

        return instructions

    @classmethod
    def read(
        cls, file_path: str, unflatten_kwargs: Optional[dict] = None, **read_kwargs
    ) -> pd.DataFrame:
        """
        Read the dataframe from the given csv file path.

        :param file_path:        The file to read the dataframe from.
        :param unflatten_kwargs: Unflatten keyword arguments for unflattening the read dataframe.
        :param read_kwargs:      Additional keyword arguments to pass to the `read_csv` function.

        :return: The loaded dataframe.
        """
        # Read the csv:
        obj = pd.read_csv(filepath_or_buffer=file_path, **read_kwargs)

        # Check if it was flattened in packing:
        if unflatten_kwargs is not None:
            # Remove the default index (joined with reset index):
            if obj.columns[0] == "Unnamed: 0":
                obj.drop(columns=["Unnamed: 0"], inplace=True)
            # Unflatten the dataframe:
            obj = cls._unflatten_dataframe(dataframe=obj, **unflatten_kwargs)

        return obj


class _H5Formatter(_Formatter):
    """
    A static class for managing pandas h5 files.
    """

    @classmethod
    def to(
        cls, obj: pd.DataFrame, file_path: str, flatten: bool = True, **to_kwargs
    ) -> dict:
        """
        Save the given dataframe to the h5 file path given.

        :param obj:       The dataframe to save.
        :param file_path: The file to save to.
        :param flatten:   Ignored for h5 format.
        :param to_kwargs: Additional keyword arguments to pass to the `to_hdf` function.

        :return A dictionary of keyword arguments for reading the dataframe from file.
        """
        # If user didn't provide a key for the dataframe, use default key 'table':
        key = to_kwargs.pop("key", "table")

        # Write to h5:
        obj.to_hdf(path_or_buf=file_path, key=key, **to_kwargs)

        return {"key": key}

    @classmethod
    def read(
        cls, file_path: str, unflatten_kwargs: Optional[dict] = None, **read_kwargs
    ) -> pd.DataFrame:
        """
        Read the dataframe from the given h5 file path.

        :param file_path:        The file to read the dataframe from.
        :param unflatten_kwargs: Ignored for h5 format.
        :param read_kwargs:      Additional keyword arguments to pass to the `read_hdf` function.

        :return: The loaded dataframe.
        """
        return pd.read_hdf(path_or_buf=file_path, **read_kwargs)


class _XMLFormatter(_Formatter):
    """
    A static class for managing pandas xml files.
    """

    @classmethod
    def to(
        cls, obj: pd.DataFrame, file_path: str, flatten: bool = True, **to_kwargs
    ) -> dict:
        """
        Save the given dataframe to the xml file path given.

        :param obj:       The dataframe to save.
        :param file_path: The file to save to.
        :param flatten:   Whether to flatten the dataframe before saving. For some formats it is mandatory to enable
                          flattening, otherwise saving and loading the dataframe will cause unexpected behavior
                          especially in case it is multi-level or multi-index. Default to True.
        :param to_kwargs: Additional keyword arguments to pass to the `to_xml` function.

        :return A dictionary of keyword arguments for reading the dataframe from file.
        """
        # Get the parser (if not provided, try to use `lxml`, otherwise `etree`):
        parser = to_kwargs.pop("parser", None)
        if parser is None:
            try:
                importlib.import_module("lxml")
                parser = "lxml"
            except ModuleNotFoundError:
                parser = "etree"
        instructions = {"parser": parser}

        # Flatten the dataframe (this format have problems saving multi-level dataframes):
        if flatten:
            obj, unflatten_kwargs = cls._flatten_dataframe(dataframe=obj)
            instructions["unflatten_kwargs"] = unflatten_kwargs

        # Write to xml:
        obj.to_xml(path_or_buffer=file_path, parser="etree", **to_kwargs)

        return instructions

    @classmethod
    def read(
        cls, file_path: str, unflatten_kwargs: Optional[dict] = None, **read_kwargs
    ) -> pd.DataFrame:
        """
        Read the dataframe from the given xml file path.

        :param file_path:        The file to read the dataframe from.
        :param unflatten_kwargs: Unflatten keyword arguments for unflattening the read dataframe.
        :param read_kwargs:      Additional keyword arguments to pass to the `read_xml` function.

        :return: The loaded dataframe.
        """
        # Read the xml:
        obj = pd.read_xml(path_or_buffer=file_path, **read_kwargs)

        # Check if it was flattened in packing:
        if unflatten_kwargs is not None:
            # Remove the default index (joined with reset index):
            if obj.columns[0] == "index":
                obj.drop(columns=["index"], inplace=True)
            # Unflatten the dataframe:
            obj = cls._unflatten_dataframe(dataframe=obj, **unflatten_kwargs)

        return obj


class _XLSXFormatter(_Formatter):
    """
    A static class for managing pandas xlsx files.
    """

    @classmethod
    def to(
        cls, obj: pd.DataFrame, file_path: str, flatten: bool = True, **to_kwargs
    ) -> dict:
        """
        Save the given dataframe to the xlsx file path given.

        :param obj:       The dataframe to save.
        :param file_path: The file to save to.
        :param flatten:   Whether to flatten the dataframe before saving. For some formats it is mandatory to enable
                          flattening, otherwise saving and loading the dataframe will cause unexpected behavior
                          especially in case it is multi-level or multi-index. Default to True.
        :param to_kwargs: Additional keyword arguments to pass to the `to_excel` function.
        """
        # Get the engine to pass when unpacked:
        instructions = {"engine": to_kwargs.get("engine", None)}

        # Flatten the dataframe (this format have problems saving multi-level dataframes):
        if flatten:
            obj, unflatten_kwargs = cls._flatten_dataframe(dataframe=obj)
            instructions["unflatten_kwargs"] = unflatten_kwargs

        # Write to xlsx:
        obj.to_excel(excel_writer=file_path, **to_kwargs)

        return instructions

    @classmethod
    def read(
        cls, file_path: str, unflatten_kwargs: Optional[dict] = None, **read_kwargs
    ) -> pd.DataFrame:
        """
        Read the dataframe from the given xlsx file path.

        :param file_path:        The file to read the dataframe from.
        :param unflatten_kwargs: Unflatten keyword arguments for unflattening the read dataframe.
        :param read_kwargs:      Additional keyword arguments to pass to the `read_excel` function.

        :return: The loaded dataframe.
        """
        # Read the xlsx:
        obj = pd.read_excel(io=file_path, **read_kwargs)

        # Check if it was flattened in packing:
        if unflatten_kwargs is not None:
            # Remove the default index (joined with reset index):
            if obj.columns[0] == "Unnamed: 0":
                obj.drop(columns=["Unnamed: 0"], inplace=True)
            # Unflatten the dataframe:
            obj = cls._unflatten_dataframe(dataframe=obj, **unflatten_kwargs)

        return obj


class _HTMLFormatter(_Formatter):
    """
    A static class for managing pandas html files.
    """

    @classmethod
    def to(
        cls, obj: pd.DataFrame, file_path: str, flatten: bool = True, **to_kwargs
    ) -> dict:
        """
        Save the given dataframe to the html file path given.

        :param obj:       The dataframe to save.
        :param file_path: The file to save to.
        :param flatten:   Whether to flatten the dataframe before saving. For some formats it is mandatory to enable
                          flattening, otherwise saving and loading the dataframe will cause unexpected behavior
                          especially in case it is multi-level or multi-index. Default to True.
        :param to_kwargs: Additional keyword arguments to pass to the `to_html` function.

        :return A dictionary of keyword arguments for reading the dataframe from file.
        """
        # Flatten the dataframe (this format have problems saving multi-level dataframes):
        instructions = {}
        if flatten:
            obj, unflatten_kwargs = cls._flatten_dataframe(dataframe=obj)
            instructions["unflatten_kwargs"] = unflatten_kwargs

        # Write to html:
        obj.to_html(buf=file_path, **to_kwargs)
        return instructions

    @classmethod
    def read(
        cls, file_path: str, unflatten_kwargs: Optional[dict] = None, **read_kwargs
    ) -> pd.DataFrame:
        """
        Read dataframes from the given html file path.

        :param file_path:        The file to read the dataframe from.
        :param unflatten_kwargs: Unflatten keyword arguments for unflattening the read dataframe.
        :param read_kwargs:      Additional keyword arguments to pass to the `read_html` function.

        :return: The loaded dataframe.
        """
        # Read the html:
        obj = pd.read_html(io=file_path, **read_kwargs)[0]

        # Check if it was flattened in packing:
        if unflatten_kwargs is not None:
            # Remove the default index (joined with reset index):
            if obj.columns[0] == "Unnamed: 0":
                obj.drop(columns=["Unnamed: 0"], inplace=True)
            # Unflatten the dataframe:
            obj = cls._unflatten_dataframe(dataframe=obj, **unflatten_kwargs)

        return obj


class _JSONFormatter(_Formatter):
    """
    A static class for managing pandas json files.
    """

    @classmethod
    def to(
        cls, obj: pd.DataFrame, file_path: str, flatten: bool = True, **to_kwargs
    ) -> dict:
        """
        Save the given dataframe to the json file path given.

        :param obj:       The dataframe to save.
        :param file_path: The file to save to.
        :param flatten:   Whether to flatten the dataframe before saving. For some formats it is mandatory to enable
                          flattening, otherwise saving and loading the dataframe will cause unexpected behavior
                          especially in case it is multi-level or multi-index. Default to True.
        :param to_kwargs: Additional keyword arguments to pass to the `to_json` function.

        :return A dictionary of keyword arguments for reading the dataframe from file.
        """
        # Get the orient to pass when unpacked:
        instructions = {"orient": to_kwargs.get("orient", None)}

        # Flatten the dataframe (this format have problems saving multi-level dataframes):
        if flatten:
            obj, unflatten_kwargs = cls._flatten_dataframe(dataframe=obj)
            instructions["unflatten_kwargs"] = unflatten_kwargs

        # Write to json:
        obj.to_json(path_or_buf=file_path, **to_kwargs)

        return instructions

    @classmethod
    def read(
        cls, file_path: str, unflatten_kwargs: Optional[dict] = None, **read_kwargs
    ) -> pd.DataFrame:
        """
        Read dataframes from the given json file path.

        :param file_path:        The file to read the dataframe from.
        :param unflatten_kwargs: Unflatten keyword arguments for unflattening the read dataframe.
        :param read_kwargs:      Additional keyword arguments to pass to the `read_json` function.

        :return: The loaded dataframe.
        """
        # Read the json:
        obj = pd.read_json(path_or_buf=file_path, **read_kwargs)

        # Check if it was flattened in packing:
        if unflatten_kwargs is not None:
            obj = cls._unflatten_dataframe(dataframe=obj, **unflatten_kwargs)

        return obj


class _FeatherFormatter(_Formatter):
    """
    A static class for managing pandas feather files.
    """

    @classmethod
    def to(
        cls, obj: pd.DataFrame, file_path: str, flatten: bool = True, **to_kwargs
    ) -> dict:
        """
        Save the given dataframe to the feather file path given.

        :param obj:       The dataframe to save.
        :param file_path: The file to save to.
        :param flatten:   Whether to flatten the dataframe before saving. For some formats it is mandatory to enable
                          flattening, otherwise saving and loading the dataframe will cause unexpected behavior
                          especially in case it is multi-level or multi-index. Default to True.
        :param to_kwargs: Additional keyword arguments to pass to the `to_feather` function.

        :return A dictionary of keyword arguments for reading the dataframe from file.
        """
        # Flatten the dataframe (this format have problems saving multi-level dataframes):
        instructions = {}
        if flatten:
            obj, unflatten_kwargs = cls._flatten_dataframe(dataframe=obj)
            instructions["unflatten_kwargs"] = unflatten_kwargs

        # Write to feather:
        obj.to_feather(path=file_path, **to_kwargs)

        return instructions

    @classmethod
    def read(
        cls, file_path: str, unflatten_kwargs: Optional[dict] = None, **read_kwargs
    ) -> pd.DataFrame:
        """
        Read dataframes from the given feather file path.

        :param file_path:        The file to read the dataframe from.
        :param unflatten_kwargs: Unflatten keyword arguments for unflattening the read dataframe.
        :param read_kwargs:      Additional keyword arguments to pass to the `read_feather` function.

        :return: The loaded dataframe.
        """
        # Read the feather:
        obj = pd.read_feather(path=file_path, **read_kwargs)

        # Check if it was flattened in packing:
        if unflatten_kwargs is not None:
            obj = cls._unflatten_dataframe(dataframe=obj, **unflatten_kwargs)

        return obj


class _ORCFormatter(_Formatter):
    """
    A static class for managing pandas orc files.
    """

    @classmethod
    def to(
        cls, obj: pd.DataFrame, file_path: str, flatten: bool = True, **to_kwargs
    ) -> dict:
        """
        Save the given dataframe to the orc file path given.

        :param obj:       The dataframe to save.
        :param file_path: The file to save to.
        :param flatten:   Whether to flatten the dataframe before saving. For some formats it is mandatory to enable
                          flattening, otherwise saving and loading the dataframe will cause unexpected behavior
                          especially in case it is multi-level or multi-index. Default to True.
        :param to_kwargs: Additional keyword arguments to pass to the `to_orc` function.

        :return A dictionary of keyword arguments for reading the dataframe from file.
        """
        # Flatten the dataframe (this format have problems saving multi-level dataframes):
        instructions = {}
        if flatten:
            obj, unflatten_kwargs = cls._flatten_dataframe(dataframe=obj)
            instructions["unflatten_kwargs"] = unflatten_kwargs

        # Write to feather:
        obj.to_orc(path=file_path, **to_kwargs)

        return instructions

    @classmethod
    def read(
        cls, file_path: str, unflatten_kwargs: Optional[dict] = None, **read_kwargs
    ) -> pd.DataFrame:
        """
        Read dataframes from the given orc file path.

        :param file_path:        The file to read the dataframe from.
        :param unflatten_kwargs: Unflatten keyword arguments for unflattening the read dataframe.
        :param read_kwargs:      Additional keyword arguments to pass to the `read_orc` function.

        :return: The loaded dataframe.
        """
        # Read the feather:
        obj = pd.read_orc(path=file_path, **read_kwargs)

        # Check if it was flattened in packing:
        if unflatten_kwargs is not None:
            obj = cls._unflatten_dataframe(dataframe=obj, **unflatten_kwargs)

        return obj


class PandasSupportedFormat(SupportedFormat[_Formatter]):
    """
    Library of Pandas formats (file extensions) supported by the Pandas packagers.
    """

    PARQUET = "parquet"
    CSV = "csv"
    H5 = "h5"
    XML = "xml"
    XLSX = "xlsx"
    HTML = "html"
    JSON = "json"
    FEATHER = "feather"
    ORC = "orc"

    _FORMAT_HANDLERS_MAP = {
        PARQUET: _ParquetFormatter,
        CSV: _CSVFormatter,
        H5: _H5Formatter,
        XML: _XMLFormatter,
        XLSX: _XLSXFormatter,
        HTML: _HTMLFormatter,
        JSON: _JSONFormatter,
        FEATHER: _FeatherFormatter,
        ORC: _ORCFormatter,
    }


# Default file formats for pandas DataFrame and Series file artifacts:
DEFAULT_PANDAS_FORMAT = PandasSupportedFormat.PARQUET
NON_STRING_COLUMN_NAMES_DEFAULT_PANDAS_FORMAT = PandasSupportedFormat.CSV


class PandasDataFramePackager(DefaultPackager):
    """
    ``pd.DataFrame`` packager.
    """

    PACKABLE_OBJECT_TYPE = pd.DataFrame
    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.DATASET

    def get_default_unpacking_artifact_type(self, data_item: DataItem) -> str:
        """
        Get the default artifact type used for unpacking. Returns dataset if the data item represents a
        `DatasetArtifact` and otherwise, file.

        :param data_item: The about to be unpacked data item.

        :return: The default artifact type.
        """
        is_artifact = data_item.get_artifact_type()
        if is_artifact and is_artifact == "datasets":
            return ArtifactType.DATASET
        return ArtifactType.FILE

    def pack_result(self, obj: pd.DataFrame, key: str) -> dict:
        """
        Pack a dataframe as a result.

        :param obj: The dataframe to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        # Parse to dictionary according to the indexes in the dataframe:
        if len(obj.index.names) > 1:
            # Multiple indexes:
            orient = "split"
        elif obj.index.name is not None:
            # Not a default index (user would likely want to keep it):
            orient = "dict"
        else:
            # Default index can be ignored:
            orient = "list"

        # Cast to dictionary:
        dataframe_dictionary = obj.to_dict(orient=orient)

        # Prepare the result (casting tuples to lists):
        dataframe_dictionary = PandasDataFramePackager._prepare_result(
            obj=dataframe_dictionary
        )

        return super().pack_result(obj=dataframe_dictionary, key=key)

    def pack_file(
        self,
        obj: pd.DataFrame,
        key: str,
        file_format: Optional[str] = None,
        flatten: bool = True,
        **to_kwargs,
    ) -> tuple[Artifact, dict]:
        """
        Pack a dataframe as a file by the given format.

        :param obj:         The series to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is parquet or csv (depends on the column names as
                            parquet cannot be used for non string column names).
        :param flatten:     Whether to flatten the dataframe before saving. For some formats it is mandatory to enable
                            flattening, otherwise saving and loading the dataframe will cause unexpected behavior
                            especially in case it is multi-level or multi-index. Default to True.
        :param to_kwargs:   Additional keyword arguments to pass to the pandas `to_x` functions.

        :return: The packed artifact and instructions.
        """
        # Set default file format if not given:
        if file_format is None:
            file_format = (
                DEFAULT_PANDAS_FORMAT
                if all(isinstance(name, str) for name in obj.columns)
                else NON_STRING_COLUMN_NAMES_DEFAULT_PANDAS_FORMAT
            )

        # Save to file:
        formatter = PandasSupportedFormat.get_format_handler(fmt=file_format)
        temp_directory = pathlib.Path(tempfile.mkdtemp())
        self.add_future_clearing_path(path=temp_directory)
        file_path = temp_directory / f"{key}.{file_format}"
        read_kwargs = formatter.to(
            obj=obj, file_path=str(file_path), flatten=flatten, **to_kwargs
        )

        # Create the artifact and instructions:
        artifact = Artifact(key=key, src_path=os.path.abspath(file_path))

        return artifact, {"file_format": file_format, "read_kwargs": read_kwargs}

    def pack_dataset(self, obj: pd.DataFrame, key: str, file_format: str = "parquet"):
        """
        Pack a pandas dataframe as a dataset.

        :param obj:         The dataframe to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is parquet.

        :return: The packed artifact and instructions.
        """
        return DatasetArtifact(key=key, df=obj, format=file_format), {}

    def unpack_file(
        self,
        data_item: DataItem,
        file_format: Optional[str] = None,
        read_kwargs: Optional[dict] = None,
    ) -> pd.DataFrame:
        """
        Unpack a pandas dataframe from file.

        :param data_item:   The data item to unpack.
        :param file_format: The file format to use for reading the series. Default is None - will be read by the file
                            extension.
        :param read_kwargs: Keyword arguments to pass to the read of the formatter.

        :return: The unpacked series.
        """
        # Get the file:
        file_path = self.get_data_item_local_path(data_item=data_item)

        # Get the archive format by the file extension if needed:
        if file_format is None:
            file_format = PandasSupportedFormat.match_format(path=file_path)
        if file_format is None:
            raise MLRunInvalidArgumentError(
                f"File format of {data_item.key} ('{''.join(pathlib.Path(file_path).suffixes)}') is not supported. "
                f"Supported formats are: {' '.join(PandasSupportedFormat.get_all_formats())}"
            )

        # Read the object:
        formatter = PandasSupportedFormat.get_format_handler(fmt=file_format)
        if read_kwargs is None:
            read_kwargs = {}
        return formatter.read(file_path=file_path, **read_kwargs)

    def unpack_dataset(self, data_item: DataItem):
        """
        Unpack a padnas dataframe from a dataset artifact.

        :param data_item: The data item to unpack.

        :return: The unpacked dataframe.
        """
        return data_item.as_df()

    @staticmethod
    def _prepare_result(obj: Union[list, dict, tuple]) -> Any:
        """
        A dataframe can be logged as a result when it being cast to a dictionary. If the dataframe has multiple indexes,
        pandas store them as a tuple, which is not json serializable, so we cast them into lists.

        :param obj: The dataframe dictionary (or list and tuple as it is recursive).

        :return: Prepared result.
        """
        if isinstance(obj, dict):
            for key, value in obj.items():
                obj[PandasDataFramePackager._prepare_result(obj=key)] = (
                    PandasDataFramePackager._prepare_result(obj=value)
                )
        elif isinstance(obj, list):
            for i, value in enumerate(obj):
                obj[i] = PandasDataFramePackager._prepare_result(obj=value)
        elif isinstance(obj, tuple):
            obj = [PandasDataFramePackager._prepare_result(obj=value) for value in obj]
        return obj


class PandasSeriesPackager(PandasDataFramePackager):
    """
    ``pd.Series`` packager.
    """

    PACKABLE_OBJECT_TYPE = pd.Series
    DEFAULT_PACKING_ARTIFACT_TYPE = ArtifactType.FILE

    def get_supported_artifact_types(self) -> list[str]:
        """
        Get all the supported artifact types on this packager. It will be the same as `PandasDataFramePackager` but
        without the 'dataset' artifact type support.

        :return: A list of all the supported artifact types.
        """
        supported_artifacts = super().get_supported_artifact_types()
        supported_artifacts.remove("dataset")
        return supported_artifacts

    def pack_result(self, obj: pd.Series, key: str) -> dict:
        """
        Pack a series as a result.

        :param obj: The series to pack and log.
        :param key: The result's key.

        :return: The result dictionary.
        """
        return super().pack_result(obj=pd.DataFrame(obj), key=key)

    def pack_file(
        self,
        obj: pd.Series,
        key: str,
        file_format: Optional[str] = None,
        flatten: bool = True,
        **to_kwargs,
    ) -> tuple[Artifact, dict]:
        """
        Pack a series as a file by the given format.

        :param obj:         The series to pack.
        :param key:         The key to use for the artifact.
        :param file_format: The file format to save as. Default is parquet or csv (depends on the column names as
                            parquet cannot be used for non string column names).
        :param flatten:     Whether to flatten the dataframe before saving. For some formats it is mandatory to enable
                            flattening, otherwise saving and loading the dataframe will cause unexpected behavior
                            especially in case it is multi-level or multi-index. Default to True.
        :param to_kwargs:   Additional keyword arguments to pass to the pandas `to_x` functions.

        :return: The packed artifact and instructions.
        """
        # Get the series column name:
        column_name = obj.name

        # Cast to dataframe and call the parent `pack_file`:
        artifact, instructions = super().pack_file(
            obj=pd.DataFrame(obj),
            key=key,
            file_format=file_format,
            flatten=flatten,
            **to_kwargs,
        )

        # Return the artifact with the updated instructions:
        return artifact, {**instructions, "column_name": column_name}

    def unpack_file(
        self,
        data_item: DataItem,
        file_format: Optional[str] = None,
        read_kwargs: Optional[dict] = None,
        column_name: Optional[Union[str, int]] = None,
    ) -> pd.Series:
        """
        Unpack a pandas series from file.

        :param data_item:     The data item to unpack.
        :param file_format:   The file format to use for reading the series. Default is None - will be read by the file
                              extension.
        :param read_kwargs:   Keyword arguments to pass to the read of the formatter.
        :param column_name:   The name of the series column.

        :return: The unpacked series.
        """
        # Read the object:
        obj = super().unpack_file(
            data_item=data_item,
            file_format=file_format,
            read_kwargs=read_kwargs,
        )

        # Cast the dataframe into a series:
        if len(obj.columns) != 1:
            raise MLRunInvalidArgumentError(
                f"The data item received is of a `pandas.DataFrame` with more than one column: "
                f"{', '.join(obj.columns)}. Hence it cannot be turned into a `pandas.Series`."
            )
        obj = obj[obj.columns[0]]

        # Edit the column name (if `read_kwargs` is not None we can be sure it is a packed file artifact, so the column
        # name, even if None, should be set to restore the object as it was):
        if read_kwargs is not None:
            obj.name = column_name

        return obj



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dependency_injector import containers, providers

import mlrun.db
import mlrun.db.httpdb
import mlrun.db.nopdb
import mlrun.utils.singleton
from mlrun.utils import logger


class RunDBFactory(
    metaclass=mlrun.utils.singleton.AbstractSingleton,
):
    def __init__(self):
        self._run_db = None
        self._last_db_url = None
        self._rundb_container = RunDBContainer()

    def create_run_db(self, url="", secrets=None, force_reconnect=False):
        """Returns the runtime database"""
        if not url:
            url = mlrun.db.get_or_set_dburl("./")

        if (
            self._last_db_url is not None
            and url == self._last_db_url
            and self._run_db
            and not force_reconnect
        ):
            return self._run_db

        self._last_db_url = url

        if "://" not in str(url):
            logger.warning(
                "Could not detect path to API server, not connected to API server!"
            )
            logger.warning(
                "MLRUN_DBPATH is misconfigured. Set this environment variable to the URL of the API server"
                " in order to connect"
            )
            self._run_db = self._rundb_container.nop(url)

        else:
            self._run_db = self._rundb_container.run_db(url)

        self._run_db.connect(secrets=secrets)
        return self._run_db


class RunDBContainer(containers.DeclarativeContainer):
    nop = providers.Factory(mlrun.db.nopdb.NopDB)
    run_db = providers.Factory(mlrun.db.httpdb.HTTPRunDB)



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import mlrun


class BaseRemoteClient:
    """
    The BaseRemoteClient class serves as a foundational component for managing
    secrets and configurations.
    It is designed to be extended by subclasses that interact with external services,
     such as file systems (e.g., Datastore) or model providers (e.g., ModelProvider).

    This class is intended to provide shared functionality and should not be
    used directly. Instead, create a subclass to implement logic specific to
    your use case, such as interactions with S3 storage or invoking model providers like OpenAI.
    """

    def __init__(self, parent, kind, name, endpoint="", secrets: Optional[dict] = None):
        self._parent = parent
        self.kind = kind
        self.name = name
        self.endpoint = endpoint
        self._secrets = secrets or {}
        self.secret_pfx = ""

    def _get_secret_or_env(self, key, default=None):
        # Project-secrets are mounted as env variables whose name can be retrieved from SecretsStore
        return mlrun.get_secret_or_env(
            key, secret_provider=self._get_secret, default=default
        )

    def _get_parent_secret(self, key):
        return self._parent.secret(self.secret_pfx + key)

    def _get_secret(self, key: str, default=None):
        return self._secrets.get(key, default) or self._get_parent_secret(key)

    @property
    def url(self):
        return f"{self.kind}://{self.endpoint}"

    @staticmethod
    def _sanitize_options(options):
        if not options:
            return {}
        options = {k: v for k, v in options.items() if v is not None and v != ""}
        return options

    @classmethod
    def parse_endpoint_and_path(cls, endpoint, subpath) -> (str, str):
        return endpoint, subpath



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import pydantic.v1

import mlrun.common.types


# Ideally we would want this to be class FunctionState(mlrun.common.types.StrEnum) which is the
# "FastAPI-compatible" way of creating schemas
# But, when we save a function to the DB, we pickle the body, which saves the state as an instance of this class (and
# not just a string), then if for some reason we downgrade to 0.6.4, before we had this class, we fail reading (pickle
# load) the function from the DB.
# Note that the problems are happening only if the state is assigned in the API side. When it's in the client side it
# anyways passes through JSON in the HTTP request body and come up as a string in the API side.
# For now I'm simply making the class a simple string consts class
# 2 other solutions I thought of:
# 1. Changing the places where we set the state in the UI to use set the actual enum value (FunctionState.x.value) - too
# fragile, tomorrow someone will set the state using the enum
# 2. Changing the function to be saved into a JSON field instead of pickled inside BLOB field - looks like the ideal we
# should go to, but too complicated and needed something fast and simple.
class FunctionState:
    unknown = "unknown"
    ready = "ready"
    error = "error"  # represents deployment error

    deploying = "deploying"
    # there is currently an abuse usage of the builder (lower) pod state as the function state, ideally these two would
    # map to deploying but for backwards compatibility reasons we have to keep them
    running = "running"
    pending = "pending"
    # same goes for the build which is not coming from the pod, but is used and we can't just omit it for BC reasons
    build = "build"

    # for pipeline steps
    skipped = "skipped"
    initialized = "initialized"

    @classmethod
    def get_function_state_from_pod_state(cls, pod_state: str):
        if pod_state == "succeeded":
            return cls.ready
        if pod_state in ["failed", "error"]:
            return cls.error
        if pod_state in ["running", "pending"]:
            return getattr(cls, pod_state)
        return cls.unknown

    @classmethod
    def terminal_states(cls):
        return [
            cls.ready,
            cls.error,
            cls.skipped,
        ]


class PreemptionModes(mlrun.common.types.StrEnum):
    # makes function pods be able to run on preemptible nodes
    allow = "allow"
    # makes the function pods run on preemptible nodes only
    constrain = "constrain"
    # prevents the function pods from running on preemptible nodes
    prevent = "prevent"
    # doesn't apply any preemptible node selection on the function
    none = "none"


# used when running in Iguazio (otherwise use disabled mode)
# populates mlrun.mlconf.function.spec.security_context.enrichment_mode
class SecurityContextEnrichmentModes(mlrun.common.types.StrEnum):
    # always use the user id of the user that triggered the 1st run / created the function
    # NOTE: this mode is incomplete and not fully supported yet
    retain = "retain"
    # use the user id of the user that triggered the current run
    override = "override"
    # security context is not auto applied
    disabled = "disabled"


class ImagePullSecret(pydantic.v1.BaseModel):
    default: typing.Optional[str]


class Pipelines(pydantic.v1.BaseModel):
    kfp_pod_user_unix_id: typing.Optional[int]


class SecurityContext(pydantic.v1.BaseModel):
    default: typing.Optional[str]
    enrichment_mode: typing.Optional[SecurityContextEnrichmentModes]
    enrichment_group_id: typing.Optional[int]
    pipelines: typing.Optional[Pipelines]


class ServiceAccount(pydantic.v1.BaseModel):
    default: typing.Optional[str]


class StateThresholds(pydantic.v1.BaseModel):
    default: typing.Optional[dict[str, str]]


class Backoff(pydantic.v1.BaseModel):
    default_base_delay: typing.Optional[str]
    min_base_delay: typing.Optional[str]


class RetrySpec(pydantic.v1.BaseModel):
    backoff: Backoff


class FunctionSpec(pydantic.v1.BaseModel):
    image_pull_secret: typing.Optional[ImagePullSecret]
    security_context: typing.Optional[SecurityContext]
    service_account: typing.Optional[ServiceAccount]
    state_thresholds: typing.Optional[StateThresholds]
    retry: typing.Optional[RetrySpec]

    class Config:
        extra = pydantic.v1.Extra.allow


class Function(pydantic.v1.BaseModel):
    spec: typing.Optional[FunctionSpec]
    application: typing.Optional[dict[str, typing.Any]]

    class Config:
        extra = pydantic.v1.Extra.allow



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math
from datetime import datetime, timedelta, timezone
from io import StringIO
from typing import Literal, Optional, Union

import pandas as pd
import v3io_frames
import v3io_frames.client

import mlrun.common.model_monitoring
import mlrun.common.schemas.model_monitoring as mm_schemas
import mlrun.feature_store.steps
import mlrun.utils.v3io_clients
from mlrun.common.schemas import EventFieldType
from mlrun.config import config
from mlrun.model_monitoring.db import TSDBConnector
from mlrun.model_monitoring.helpers import get_invocations_fqn, get_start_end
from mlrun.utils import logger

_TSDB_BE = "tsdb"
_TSDB_RATE = "1/s"
_CONTAINER = "users"

V3IO_FRAMESD_MEPS_LIMIT = (
    200  # Maximum number of model endpoints per single request when using V3IO Frames
)
V3IO_CLIENT_MEPS_LIMIT = (
    150  # Maximum number of model endpoints per single request when using V3IO Client
)


def _is_no_schema_error(exc: v3io_frames.Error) -> bool:
    """
    In case of a nonexistent TSDB table - a `v3io_frames.ReadError` error is raised.
    Check if the error message contains the relevant string to verify the cause.
    """
    msg = str(exc)
    # https://github.com/v3io/v3io-tsdb/blob/v0.14.1/pkg/tsdb/v3iotsdb.go#L205
    # https://github.com/v3io/v3io-tsdb/blob/v0.14.1/pkg/partmgr/partmgr.go#L238
    return "No TSDB schema file found" in msg or "Failed to read schema at path" in msg


class V3IOTSDBConnector(TSDBConnector):
    """
    Handles the TSDB operations when the TSDB connector is of type V3IO. To manage these operations we use V3IO Frames
    Client that provides API for executing commands on the V3IO TSDB table.
    """

    type: str = mm_schemas.TSDBTarget.V3IO_TSDB

    def __init__(
        self,
        project: str,
        container: str = _CONTAINER,
        v3io_framesd: Optional[str] = None,
        v3io_access_key: str = "",
        create_table: bool = False,
    ) -> None:
        super().__init__(project=project)

        self.container = container

        self.v3io_framesd = v3io_framesd or mlrun.mlconf.v3io_framesd
        self._v3io_access_key = v3io_access_key
        self._frames_client: Optional[v3io_frames.client.ClientBase] = None
        self._init_tables_path()
        self._create_table = create_table
        self._v3io_client = None

    @property
    def v3io_client(self):
        if not self._v3io_client:
            self._v3io_client = mlrun.utils.v3io_clients.get_v3io_client(
                endpoint=mlrun.mlconf.v3io_api, access_key=self._v3io_access_key
            )
        return self._v3io_client

    @property
    def frames_client(self) -> v3io_frames.client.ClientBase:
        if not self._frames_client:
            self._frames_client = self._get_v3io_frames_client(
                self.container, v3io_access_key=self._v3io_access_key
            )
            if self._create_table:
                self.create_tables()
        return self._frames_client

    def _init_tables_path(self):
        self.tables = {}

        events_table_full_path = mlrun.mlconf.get_model_monitoring_file_target_path(
            project=self.project,
            kind=mm_schemas.FileTargetKind.EVENTS,
        )
        (
            _,
            _,
            events_path,
        ) = mlrun.common.model_monitoring.helpers.parse_model_endpoint_store_prefix(
            events_table_full_path
        )
        self.tables[mm_schemas.V3IOTSDBTables.EVENTS] = events_path

        errors_table_full_path = mlrun.mlconf.get_model_monitoring_file_target_path(
            project=self.project,
            kind=mm_schemas.FileTargetKind.ERRORS,
        )
        (
            _,
            _,
            errors_path,
        ) = mlrun.common.model_monitoring.helpers.parse_model_endpoint_store_prefix(
            errors_table_full_path
        )
        self.tables[mm_schemas.V3IOTSDBTables.ERRORS] = errors_path

        monitoring_application_full_path = (
            mlrun.mlconf.get_model_monitoring_file_target_path(
                project=self.project,
                kind=mm_schemas.FileTargetKind.MONITORING_APPLICATION,
            )
        )
        (
            _,
            _,
            monitoring_application_path,
        ) = mlrun.common.model_monitoring.helpers.parse_model_endpoint_store_prefix(
            monitoring_application_full_path
        )
        self.tables[mm_schemas.V3IOTSDBTables.APP_RESULTS] = (
            monitoring_application_path + mm_schemas.V3IOTSDBTables.APP_RESULTS
        )
        self.tables[mm_schemas.V3IOTSDBTables.METRICS] = (
            monitoring_application_path + mm_schemas.V3IOTSDBTables.METRICS
        )

        monitoring_predictions_full_path = (
            mlrun.mlconf.get_model_monitoring_file_target_path(
                project=self.project,
                kind=mm_schemas.V3IOTSDBTables.PREDICTIONS,
            )
        )
        (
            _,
            _,
            monitoring_predictions_path,
        ) = mlrun.common.model_monitoring.helpers.parse_model_endpoint_store_prefix(
            monitoring_predictions_full_path
        )
        self.tables[mm_schemas.V3IOTSDBTables.PREDICTIONS] = monitoring_predictions_path

        # initialize kv table
        last_request_full_table_path = (
            mlrun.mlconf.get_model_monitoring_file_target_path(
                project=self.project,
                kind=mm_schemas.FileTargetKind.LAST_REQUEST,
            )
        )
        (
            _,
            _,
            self.last_request_table,
        ) = mlrun.common.model_monitoring.helpers.parse_model_endpoint_store_prefix(
            last_request_full_table_path
        )

    def create_tables(self) -> None:
        """
        Create the tables using the TSDB connector. These are the tables that are stored in the V3IO TSDB:
        - app_results: a detailed result that includes status, kind, extra data, etc.
        - metrics: a basic key value that represents a single numeric metric.
        - events: A statistics table that includes pre-aggregated metrics (such as average latency over the
        last 5 minutes) and data samples
        - predictions: a detailed prediction that includes latency, request timestamp, etc. This table also
        includes pre-aggregated operations such as count and average on 1 minute granularity.
        - errors: a detailed error that includes error desc, error type, etc.

        """

        default_configurations = {
            "backend": _TSDB_BE,
            "if_exists": v3io_frames.IGNORE,
            "rate": _TSDB_RATE,
        }

        for table_name in self.tables:
            default_configurations["table"] = self.tables[table_name]
            if table_name == mm_schemas.V3IOTSDBTables.PREDICTIONS:
                default_configurations["aggregates"] = "count,avg"
                default_configurations["aggregation_granularity"] = "1m"
            elif table_name == mm_schemas.V3IOTSDBTables.EVENTS:
                default_configurations["rate"] = "10/m"
            logger.info("Creating table in V3IO TSDB", table_name=table_name)
            self.frames_client.create(**default_configurations)

    def apply_monitoring_stream_steps(
        self,
        graph,
        tsdb_batching_max_events: int = 1000,
        tsdb_batching_timeout_secs: int = 30,
        sample_window: int = 10,
        aggregate_windows: Optional[list[str]] = None,
        aggregate_period: str = "1m",
        **kwarg,
    ):
        """
        Apply TSDB steps on the provided monitoring graph. Throughout these steps, the graph stores live data of
        different key metric dictionaries.This data is being used by the monitoring dashboards in
        grafana. Results can be found under  v3io:///users/pipelines/project-name/model-endpoints/events/.
        In that case, we generate 3 different key  metric dictionaries:
        - base_metrics (average latency and predictions over time)
        - endpoint_features (Prediction and feature names and values)
        - custom_metrics (user-defined metrics)
        """
        aggregate_windows = aggregate_windows or ["5m", "1h"]

        # Calculate number of predictions and average latency
        def apply_storey_aggregations():
            # Calculate number of predictions for each window (5 min and 1 hour by default)
            graph.add_step(
                class_name="storey.AggregateByKey",
                aggregates=[
                    {
                        "name": EventFieldType.LATENCY,
                        "column": EventFieldType.LATENCY,
                        "operations": ["count", "avg"],
                        "windows": aggregate_windows,
                        "period": aggregate_period,
                    }
                ],
                name=EventFieldType.LATENCY,
                after="FilterNOP",
                step_name="Aggregates",
                table=".",
                key_field=EventFieldType.ENDPOINT_ID,
            )
            # Calculate average latency time for each window (5 min and 1 hour by default)
            graph.add_step(
                class_name="storey.Rename",
                mapping={
                    "latency_count_5m": mm_schemas.EventLiveStats.PREDICTIONS_COUNT_5M,
                    "latency_count_1h": mm_schemas.EventLiveStats.PREDICTIONS_COUNT_1H,
                },
                name="Rename",
                after=EventFieldType.LATENCY,
            )

        apply_storey_aggregations()
        # Write latency per prediction, labeled by endpoint ID only
        graph.add_step(
            "storey.TSDBTarget",
            name="tsdb_predictions",
            after="FilterNOP",
            path=f"{self.container}/{self.tables[mm_schemas.V3IOTSDBTables.PREDICTIONS]}",
            time_col=mm_schemas.EventFieldType.TIMESTAMP,
            container=self.container,
            v3io_frames=self.v3io_framesd,
            columns=[
                mm_schemas.EventFieldType.LATENCY,
                mm_schemas.EventFieldType.LAST_REQUEST_TIMESTAMP,
                mm_schemas.EventFieldType.ESTIMATED_PREDICTION_COUNT,
                mm_schemas.EventFieldType.EFFECTIVE_SAMPLE_COUNT,
            ],
            index_cols=[
                mm_schemas.EventFieldType.ENDPOINT_ID,
            ],
            max_events=tsdb_batching_max_events,
            flush_after_seconds=tsdb_batching_timeout_secs,
            key=mm_schemas.EventFieldType.ENDPOINT_ID,
        )

        # Write last request timestamp to KV table
        graph.add_step(
            "storey.NoSqlTarget",
            name="KVLastRequest",
            after="tsdb_predictions",
            table=f"v3io:///users/{self.last_request_table}",
            columns=[EventFieldType.LAST_REQUEST_TIMESTAMP],
            index_cols=[EventFieldType.ENDPOINT_ID],
        )

        # Emits the event in window size of events based on sample_window size (10 by default)
        graph.add_step(
            "storey.steps.SampleWindow",
            name="sample",
            after="Rename",
            window_size=sample_window,
            key=EventFieldType.ENDPOINT_ID,
        )

        # Before writing data to TSDB, create dictionary of 2-3 dictionaries that contains
        # stats and details about the events

        graph.add_step(
            "mlrun.model_monitoring.db.tsdb.v3io.stream_graph_steps.ProcessBeforeTSDB",
            name="ProcessBeforeTSDB",
            after="sample",
        )

        # Unpacked keys from each dictionary and write to TSDB target
        def apply_filter_and_unpacked_keys(name, keys):
            graph.add_step(
                "mlrun.model_monitoring.db.tsdb.v3io.stream_graph_steps.FilterAndUnpackKeys",
                name=name,
                after="ProcessBeforeTSDB",
                keys=[keys],
            )

        def apply_tsdb_target(name, after):
            graph.add_step(
                "storey.TSDBTarget",
                name=name,
                after=after,
                path=f"{self.container}/{self.tables[mm_schemas.V3IOTSDBTables.EVENTS]}",
                time_col=mm_schemas.EventFieldType.TIMESTAMP,
                container=self.container,
                v3io_frames=self.v3io_framesd,
                infer_columns_from_data=True,
                index_cols=[
                    mm_schemas.EventFieldType.ENDPOINT_ID,
                    mm_schemas.EventFieldType.RECORD_TYPE,
                    mm_schemas.EventFieldType.ENDPOINT_TYPE,
                ],
                max_events=tsdb_batching_max_events,
                flush_after_seconds=tsdb_batching_timeout_secs,
                key=mm_schemas.EventFieldType.ENDPOINT_ID,
            )

        # unpacked base_metrics dictionary
        apply_filter_and_unpacked_keys(
            name="FilterAndUnpackKeys1",
            keys=mm_schemas.EventKeyMetrics.BASE_METRICS,
        )
        apply_tsdb_target(name="tsdb1", after="FilterAndUnpackKeys1")

        # unpacked endpoint_features dictionary
        apply_filter_and_unpacked_keys(
            name="FilterAndUnpackKeys2",
            keys=mm_schemas.EventKeyMetrics.ENDPOINT_FEATURES,
        )
        apply_tsdb_target(name="tsdb2", after="FilterAndUnpackKeys2")

        # unpacked custom_metrics dictionary. In addition, use storey.Filter remove none values
        apply_filter_and_unpacked_keys(
            name="FilterAndUnpackKeys3",
            keys=mm_schemas.EventKeyMetrics.CUSTOM_METRICS,
        )

        def apply_storey_filter():
            graph.add_step(
                "storey.Filter",
                "FilterNotNone",
                after="FilterAndUnpackKeys3",
                _fn="(event is not None)",
            )

        apply_storey_filter()
        apply_tsdb_target(name="tsdb3", after="FilterNotNone")

    def apply_writer_steps(self, graph, after, **kwargs) -> None:
        graph.add_step(
            "storey.TSDBTarget",
            name="tsdb_metrics",
            after=after,
            path=f"{self.container}/{self.tables[mm_schemas.V3IOTSDBTables.METRICS]}",
            time_col=mm_schemas.WriterEvent.END_INFER_TIME,
            container=self.container,
            v3io_frames=self.v3io_framesd,
            infer_columns_from_data=True,
            graph_shape="cylinder",
            index_cols=[
                mm_schemas.WriterEvent.APPLICATION_NAME,
                mm_schemas.WriterEvent.ENDPOINT_NAME,
                mm_schemas.WriterEvent.ENDPOINT_ID,
                mm_schemas.MetricData.METRIC_NAME,
            ],
            max_events=config.model_endpoint_monitoring.writer_graph.max_events,
            flush_after_seconds=config.model_endpoint_monitoring.writer_graph.flush_after_seconds,
            key=mm_schemas.EventFieldType.ENDPOINT_ID,
        )

        graph.add_step(
            "storey.TSDBTarget",
            name="tsdb_app_results",
            after=after,
            path=f"{self.container}/{self.tables[mm_schemas.V3IOTSDBTables.APP_RESULTS]}",
            time_col=mm_schemas.WriterEvent.END_INFER_TIME,
            container=self.container,
            v3io_frames=self.v3io_framesd,
            infer_columns_from_data=True,
            graph_shape="cylinder",
            index_cols=[
                mm_schemas.WriterEvent.APPLICATION_NAME,
                mm_schemas.WriterEvent.ENDPOINT_NAME,
                mm_schemas.WriterEvent.ENDPOINT_ID,
                mm_schemas.ResultData.RESULT_NAME,
            ],
            max_events=config.model_endpoint_monitoring.writer_graph.max_events,
            flush_after_seconds=config.model_endpoint_monitoring.writer_graph.flush_after_seconds,
            key=mm_schemas.EventFieldType.ENDPOINT_ID,
        )

    def handle_model_error(
        self,
        graph,
        tsdb_batching_max_events: int = 1000,
        tsdb_batching_timeout_secs: int = 30,
        **kwargs,
    ) -> None:
        graph.add_step(
            "mlrun.model_monitoring.db.tsdb.v3io.stream_graph_steps.ErrorExtractor",
            name="error_extractor",
            after="ForwardError",
        )

        graph.add_step(
            "storey.TSDBTarget",
            name="tsdb_error",
            after="error_extractor",
            path=f"{self.container}/{self.tables[mm_schemas.FileTargetKind.ERRORS]}",
            time_col=mm_schemas.EventFieldType.TIMESTAMP,
            container=self.container,
            v3io_frames=self.v3io_framesd,
            columns=[
                mm_schemas.EventFieldType.MODEL_ERROR,
                mm_schemas.EventFieldType.ERROR_COUNT,
            ],
            index_cols=[
                mm_schemas.EventFieldType.ENDPOINT_ID,
                mm_schemas.EventFieldType.ERROR_TYPE,
            ],
            max_events=tsdb_batching_max_events,
            flush_after_seconds=tsdb_batching_timeout_secs,
            key=mm_schemas.EventFieldType.ENDPOINT_ID,
        )

    def write_application_event(
        self,
        event: dict,
        kind: mm_schemas.WriterEventKind = mm_schemas.WriterEventKind.RESULT,
    ) -> None:
        """Write a single result or metric to TSDB"""

        event[mm_schemas.WriterEvent.END_INFER_TIME] = datetime.fromisoformat(
            event[mm_schemas.WriterEvent.END_INFER_TIME]
        )
        index_cols_base = [
            mm_schemas.WriterEvent.END_INFER_TIME,
            mm_schemas.WriterEvent.ENDPOINT_ID,
            mm_schemas.WriterEvent.APPLICATION_NAME,
            mm_schemas.WriterEvent.ENDPOINT_NAME,
        ]

        if kind == mm_schemas.WriterEventKind.METRIC:
            table = self.tables[mm_schemas.V3IOTSDBTables.METRICS]
            index_cols = index_cols_base + [mm_schemas.MetricData.METRIC_NAME]
        elif kind == mm_schemas.WriterEventKind.RESULT:
            table = self.tables[mm_schemas.V3IOTSDBTables.APP_RESULTS]
            index_cols = index_cols_base + [mm_schemas.ResultData.RESULT_NAME]
        else:
            raise ValueError(f"Invalid {kind = }")

        try:
            self.frames_client.write(
                backend=_TSDB_BE,
                table=table,
                dfs=pd.DataFrame.from_records([event]),
                index_cols=index_cols,
            )
            logger.info("Updated V3IO TSDB successfully", table=table)
        except v3io_frames.Error as err:
            logger.exception(
                "Could not write drift measures to TSDB",
                err=err,
                table=table,
                event=event,
            )
            raise mlrun.errors.MLRunRuntimeError(
                f"Failed to write application result to TSDB: {err}"
            )

    def delete_tsdb_resources(self, table: Optional[str] = None):
        if table:
            # Delete a specific table
            tables = [table]
        else:
            # Delete all tables
            tables = mm_schemas.V3IOTSDBTables.list()
        for table_to_delete in tables:
            if table_to_delete in self.tables:
                try:
                    self.frames_client.delete(
                        backend=_TSDB_BE, table=self.tables[table_to_delete]
                    )
                except v3io_frames.DeleteError as e:
                    logger.warning(
                        f"Failed to delete TSDB table '{table_to_delete}'",
                        err=mlrun.errors.err_to_str(e),
                    )
            else:
                logger.warning(
                    f"Skipping deletion: table '{table_to_delete}' is not among the initialized tables.",
                    initialized_tables=list(self.tables.keys()),
                )

        # Final cleanup of tsdb path
        tsdb_path = self._get_v3io_source_directory()
        tsdb_path.replace("://u", ":///u")
        store, _, _ = mlrun.store_manager.get_or_create_store(tsdb_path)
        store.rm(tsdb_path, recursive=True)

    def delete_tsdb_records(
        self,
        endpoint_ids: list[str],
    ):
        logger.debug(
            "Deleting model endpoints resources using the V3IO TSDB connector",
            project=self.project,
            number_of_endpoints_to_delete=len(endpoint_ids),
        )
        tables = mm_schemas.V3IOTSDBTables.list()

        # Split the endpoint ids into chunks to avoid exceeding the v3io-engine filter-expression limit
        for i in range(0, len(endpoint_ids), V3IO_FRAMESD_MEPS_LIMIT):
            endpoint_id_chunk = endpoint_ids[i : i + V3IO_FRAMESD_MEPS_LIMIT]
            endpoints_list = "', '".join(endpoint_id_chunk)
            filter_query = f"endpoint_id IN('{endpoints_list}')"
            for table in tables:
                try:
                    self.frames_client.delete(
                        backend=_TSDB_BE,
                        table=self.tables[table],
                        filter=filter_query,
                        start="0",
                    )
                except Exception as e:
                    logger.warning(
                        f"Failed to delete TSDB records for the provided endpoints from table '{table}'",
                        error=mlrun.errors.err_to_str(e),
                        project=self.project,
                    )

        # Clean the last request records from the KV table
        self._delete_last_request_records(endpoint_ids=endpoint_ids)

        logger.debug(
            "Deleted all model endpoint resources using the V3IO connector",
            project=self.project,
            number_of_endpoints_to_delete=len(endpoint_ids),
        )

    def _delete_last_request_records(self, endpoint_ids: list[str]):
        for endpoint_id in endpoint_ids:
            try:
                self.v3io_client.kv.delete(
                    container=self.container,
                    table_path=self.last_request_table,
                    key=endpoint_id,
                )
            except Exception as e:
                logger.warning(
                    f"Failed to delete last request record for endpoint '{endpoint_id}'",
                    error=mlrun.errors.err_to_str(e),
                    project=self.project,
                )

    def delete_application_records(
        self, application_name: str, endpoint_ids: Optional[list[str]] = None
    ) -> None:
        """
        Delete application records from the TSDB for the given model endpoints or all if ``endpoint_ids`` is ``None``.
        """
        base_filter_query = f"application_name=='{application_name}'"

        filter_queries: list[str] = []
        if endpoint_ids:
            for i in range(0, len(endpoint_ids), V3IO_FRAMESD_MEPS_LIMIT):
                endpoint_id_chunk = endpoint_ids[i : i + V3IO_FRAMESD_MEPS_LIMIT]
                endpoints_list = "', '".join(endpoint_id_chunk)
                filter_queries.append(
                    f"{base_filter_query} AND endpoint_id IN ('{endpoints_list}')"
                )
        else:
            filter_queries = [base_filter_query]

        for table in [
            self.tables[mm_schemas.V3IOTSDBTables.APP_RESULTS],
            self.tables[mm_schemas.V3IOTSDBTables.METRICS],
        ]:
            logger.debug(
                "Deleting application records from TSDB",
                table=table,
                filter_queries=filter_queries,
                project=self.project,
            )
            for filter_query in filter_queries:
                self.frames_client.delete(
                    backend=_TSDB_BE,
                    table=table,
                    filter=filter_query,
                    start="0",
                )

    def get_model_endpoint_real_time_metrics(
        self, endpoint_id: str, metrics: list[str], start: str, end: str
    ) -> dict[str, list[tuple[str, float]]]:
        """
        Getting real time metrics from the TSDB. There are pre-defined metrics for model endpoints such as
        `predictions_per_second` and `latency_avg_5m` but also custom metrics defined by the user. Note that these
        metrics are being calculated by the model monitoring stream pod.
        :param endpoint_id:      The unique id of the model endpoint.
        :param metrics:          A list of real-time metrics to return for the model endpoint.
        :param start:            The start time of the metrics. Can be represented by a string containing an RFC 3339
                                 time, a Unix timestamp in milliseconds, a relative time (`'now'` or
                                 `'now-[0-9]+[mhd]'`, where `m` = minutes, `h` = hours, `'d'` = days, and
                                 `'s'` = seconds), or 0 for the earliest time.
        :param end:              The end time of the metrics. Can be represented by a string containing an RFC 3339
                                 time, a Unix timestamp in milliseconds, a relative time (`'now'` or
                                 `'now-[0-9]+[mhd]'`, where `m` = minutes, `h` = hours, and `'d'` = days, and
                                 `'s'` = seconds), or 0 for the earliest time.
        :return: A dictionary of metrics in which the key is a metric name and the value is a list of tuples that
                 includes timestamps and the values.
        """

        if not metrics:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Metric names must be provided"
            )

        metrics_mapping = {}

        try:
            data = self._get_records(
                table=mm_schemas.V3IOTSDBTables.EVENTS,
                columns=["endpoint_id", *metrics],
                filter_query=f"endpoint_id=='{endpoint_id}'",
                start=start,
                end=end,
            )

            # Fill the metrics mapping dictionary with the metric name and values
            data_dict = data.to_dict()
            for metric in metrics:
                metric_data = data_dict.get(metric)
                if metric_data is None:
                    continue

                values = [
                    (str(timestamp), value) for timestamp, value in metric_data.items()
                ]
                metrics_mapping[metric] = values

        except v3io_frames.Error as err:
            logger.warn("Failed to read tsdb", err=err, endpoint=endpoint_id)

        return metrics_mapping

    def _get_records(
        self,
        table: str,
        start: Union[datetime, str],
        end: Union[datetime, str],
        columns: Optional[list[str]] = None,
        filter_query: str = "",
        interval: Optional[str] = None,
        agg_funcs: Optional[list[str]] = None,
        sliding_window_step: Optional[str] = None,
        get_raw: bool = False,
        **kwargs,
    ) -> Union[pd.DataFrame, list[v3io_frames.client.RawFrame]]:
        """
         Getting records from V3IO TSDB data collection.
        :param table:                 Path to the collection to query.
        :param start:                 The start time of the metrics. Can be represented by a string containing an RFC
                                      3339 time, a Unix timestamp in milliseconds, a relative time (`'now'` or
                                      `'now-[0-9]+[mhd]'`, where `m` = minutes, `h` = hours, `'d'` = days, and
                                      `'s'` = seconds), or 0 for the earliest time.
        :param end:                   The end time of the metrics. Can be represented by a string containing an RFC
                                      3339 time, a Unix timestamp in milliseconds, a relative time (`'now'` or
                                      `'now-[0-9]+[mhd]'`, where `m` = minutes, `h` = hours, `'d'` = days, and
                                      `'s'` = seconds), or 0 for the earliest time.
        :param columns:               Columns to include in the result.
        :param filter_query:          V3IO filter expression. The expected filter expression includes different
                                      conditions, divided by ' AND '.
        :param interval:              The interval to aggregate the data by. Note that if interval is provided,
                                      agg_funcs must bg provided as well. Provided as a string in the format of '1m',
                                      '1h', etc.
        :param agg_funcs:             The aggregation functions to apply on the columns. Note that if `agg_funcs` is
                                      provided, `interval` must bg provided as well. Provided as a list of strings in
                                      the format of ['sum', 'avg', 'count', ...].
        :param sliding_window_step:   The time step for which the time window moves forward. Note that if
                                      `sliding_window_step` is provided, interval must be provided as well. Provided
                                      as a string in the format of '1m', '1h', etc.
        :param kwargs:                Additional keyword arguments passed to the read method of frames client.
        :param get_raw:               Whether to return the request as raw frames rather than a pandas dataframe.
                                      Defaults to False. This can greatly improve performance when a dataframe isn't
                                      needed.

        :return: DataFrame with the provided attributes from the data collection.
        :raise:  MLRunNotFoundError if the provided table wasn't found.
        """
        if table not in self.tables:
            raise mlrun.errors.MLRunNotFoundError(
                f"Table '{table}' does not exist in the tables list of the TSDB connector. "
                f"Available tables: {list(self.tables.keys())}"
            )

        # Frames client expects the aggregators to be a comma-separated string
        aggregators = ",".join(agg_funcs) if agg_funcs else None
        table_path = self.tables[table]
        try:
            res = self.frames_client.read(
                backend=_TSDB_BE,
                table=table_path,
                start=start,
                end=end,
                columns=columns,
                filter=filter_query,
                aggregation_window=interval,
                aggregators=aggregators,
                step=sliding_window_step,
                get_raw=get_raw,
                **kwargs,
            )
            if get_raw:
                res = list(res)
        except v3io_frames.Error as err:
            if _is_no_schema_error(err):
                return [] if get_raw else pd.DataFrame()
            else:
                raise err

        return res

    def _get_v3io_source_directory(self) -> str:
        """
        Get the V3IO source directory for the current project. Usually the source directory will
        be under 'v3io:///users/pipelines/<project>'

        :return: The V3IO source directory for the current project.
        """
        events_table_full_path = mlrun.mlconf.get_model_monitoring_file_target_path(
            project=self.project,
            kind=mm_schemas.FileTargetKind.EVENTS,
        )

        # Generate the main directory with the V3IO resources
        source_directory = (
            mlrun.common.model_monitoring.helpers.parse_model_endpoint_project_prefix(
                events_table_full_path, self.project
            )
        )

        return source_directory

    @staticmethod
    def _get_v3io_frames_client(
        v3io_container: str, v3io_access_key: str = ""
    ) -> v3io_frames.client.ClientBase:
        return mlrun.utils.v3io_clients.get_frames_client(
            address=mlrun.mlconf.v3io_framesd,
            container=v3io_container,
            token=v3io_access_key,
        )

    @staticmethod
    def _generate_filter_query(
        filter_key: str, filter_values: Union[str, list[str]]
    ) -> Optional[str]:
        if isinstance(filter_values, str):
            return f"{filter_key}=='{filter_values}'"
        elif isinstance(filter_values, list):
            if len(filter_values) > V3IO_FRAMESD_MEPS_LIMIT:
                logger.info(
                    "The number of filter values exceeds the v3io-engine filter-expression limit, "
                    "retrieving all the values from the db.",
                    filter_key=filter_key,
                    limit=V3IO_FRAMESD_MEPS_LIMIT,
                    amount=len(filter_values),
                )
                return None
            return f"{filter_key} IN ({', '.join(repr(v) for v in filter_values)}) "
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Invalid filter key {filter_key}: must be a string or a list, got {type(filter_values).__name__}; "
                f"filter values: {filter_values}"
            )

    def read_metrics_data(
        self,
        *,
        endpoint_id: str,
        start: datetime,
        end: datetime,
        metrics: list[mm_schemas.ModelEndpointMonitoringMetric],
        type: Literal["metrics", "results"] = "results",
        with_result_extra_data: bool = False,
    ) -> Union[
        list[
            Union[
                mm_schemas.ModelEndpointMonitoringResultValues,
                mm_schemas.ModelEndpointMonitoringMetricNoData,
            ],
        ],
        list[
            Union[
                mm_schemas.ModelEndpointMonitoringMetricValues,
                mm_schemas.ModelEndpointMonitoringMetricNoData,
            ],
        ],
    ]:
        """
        Read metrics OR results from the TSDB and return as a list.
        Note: the type must match the actual metrics in the `metrics` parameter.
        If the type is "results", pass only results in the `metrics` parameter.
        """

        if type == "metrics":
            if with_result_extra_data:
                logger.warning(
                    "The 'with_result_extra_data' parameter is not supported for metrics, just for results",
                    project=self.project,
                    endpoint_id=endpoint_id,
                )
            table_path = self.tables[mm_schemas.V3IOTSDBTables.METRICS]
            name = mm_schemas.MetricData.METRIC_NAME
            columns = [mm_schemas.MetricData.METRIC_VALUE]
            df_handler = self.df_to_metrics_values
        elif type == "results":
            table_path = self.tables[mm_schemas.V3IOTSDBTables.APP_RESULTS]
            name = mm_schemas.ResultData.RESULT_NAME
            columns = [
                mm_schemas.ResultData.RESULT_VALUE,
                mm_schemas.ResultData.RESULT_STATUS,
                mm_schemas.ResultData.RESULT_KIND,
            ]
            if with_result_extra_data:
                columns.append(mm_schemas.ResultData.RESULT_EXTRA_DATA)
            df_handler = self.df_to_results_values
        else:
            raise ValueError(f"Invalid {type = }")

        query = self._get_sql_query(
            endpoint_id=endpoint_id,
            metric_and_app_names=[(metric.app, metric.name) for metric in metrics],
            table_path=table_path,
            name=name,
            columns=columns,
        )

        logger.debug("Querying V3IO TSDB", query=query)

        df: pd.DataFrame = self.frames_client.read(
            backend=_TSDB_BE,
            start=start,
            end=end,
            query=query,  # the filter argument does not work for this complex condition
        )

        logger.debug(
            "Converting a DataFrame to a list of metrics or results values",
            table=table_path,
            project=self.project,
            endpoint_id=endpoint_id,
            is_empty=df.empty,
        )
        if not with_result_extra_data and type == "results":
            # Set the extra data to an empty string if it's not requested
            df[mm_schemas.ResultData.RESULT_EXTRA_DATA] = ""

        return df_handler(df=df, metrics=metrics, project=self.project)

    @staticmethod
    def _get_sql_query(
        *,
        table_path: str,
        endpoint_id: Optional[str] = None,
        application_names: Optional[list[str]] = None,
        name: str = mm_schemas.ResultData.RESULT_NAME,
        metric_and_app_names: Optional[list[tuple[str, str]]] = None,
        columns: Optional[list[str]] = None,
        group_by_columns: Optional[list[str]] = None,
    ) -> str:
        """Get the SQL query for the results/metrics table"""

        if metric_and_app_names and not endpoint_id:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "If metric_and_app_names is provided, endpoint_id must also be provided"
            )

        if metric_and_app_names and application_names:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Cannot provide both metric_and_app_names and application_names"
            )

        if columns:
            selection = ",".join(columns)
        else:
            selection = "*"

        with StringIO() as query:
            where_added = False
            query.write(f"SELECT {selection} FROM '{table_path}'")
            if endpoint_id:
                query.write(
                    f" WHERE {mm_schemas.WriterEvent.ENDPOINT_ID}='{endpoint_id}'"
                )
                where_added = True
            if metric_and_app_names:
                if where_added:
                    query.write(" AND (")
                else:
                    query.write(" WHERE (")
                    where_added = True

                for i, (app_name, result_name) in enumerate(metric_and_app_names):
                    sub_cond = (
                        f"({mm_schemas.WriterEvent.APPLICATION_NAME}='{app_name}' "
                        f"AND {name}='{result_name}')"
                    )
                    if i != 0:  # not first sub condition
                        query.write(" OR ")
                    query.write(sub_cond)

                query.write(")")

            if application_names:
                if where_added:
                    query.write(" AND (")
                else:
                    query.write(" WHERE (")
                for i, app_name in enumerate(application_names):
                    sub_cond = f"{mm_schemas.WriterEvent.APPLICATION_NAME}='{app_name}'"
                    if i != 0:  # not first sub condition
                        query.write(" OR ")
                    query.write(sub_cond)
                query.write(")")

            if group_by_columns:
                query.write(" GROUP BY ")
                query.write(",".join(group_by_columns))

            query.write(";")
            return query.getvalue()

    def read_predictions(
        self,
        *,
        endpoint_id: str,
        start: Union[datetime, str],
        end: Union[datetime, str],
        aggregation_window: Optional[str] = None,
        agg_funcs: Optional[list[str]] = None,
        limit: Optional[
            int
        ] = None,  # no effect, just for compatibility with the abstract method
    ) -> Union[
        mm_schemas.ModelEndpointMonitoringMetricNoData,
        mm_schemas.ModelEndpointMonitoringMetricValues,
    ]:
        if (agg_funcs and not aggregation_window) or (
            aggregation_window and not agg_funcs
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "both or neither of `aggregation_window` and `agg_funcs` must be provided"
            )
        df = self._get_records(
            table=mm_schemas.V3IOTSDBTables.PREDICTIONS,
            start=start,
            end=end,
            columns=[mm_schemas.EventFieldType.ESTIMATED_PREDICTION_COUNT],
            filter_query=f"endpoint_id=='{endpoint_id}'",
            agg_funcs=agg_funcs,
            sliding_window_step=aggregation_window,
        )

        full_name = get_invocations_fqn(self.project)

        if df.empty:
            return mm_schemas.ModelEndpointMonitoringMetricNoData(
                full_name=full_name,
                type=mm_schemas.ModelEndpointMonitoringMetricType.METRIC,
            )

        estimated_prediction_count = (
            f"{agg_funcs[0]}({mm_schemas.EventFieldType.ESTIMATED_PREDICTION_COUNT})"
            if agg_funcs
            else mm_schemas.EventFieldType.ESTIMATED_PREDICTION_COUNT
        )

        return mm_schemas.ModelEndpointMonitoringMetricValues(
            full_name=full_name,
            values=list(
                zip(
                    df.index,
                    df[estimated_prediction_count],
                )
            ),  # pyright: ignore[reportArgumentType]
        )

    def get_last_request(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> dict[str, float]:
        if not endpoint_ids:
            return {}

        # Get the last request timestamp for each endpoint from the KV table.
        # The result of the query is a list of dictionaries,
        # each dictionary contains the endpoint id and the last request timestamp.
        last_request_timestamps = {}
        if isinstance(endpoint_ids, str):
            endpoint_ids = [endpoint_ids]

        try:
            if len(endpoint_ids) > V3IO_CLIENT_MEPS_LIMIT:
                logger.warning(
                    "The number of endpoint ids exceeds the v3io-engine filter-expression limit, "
                    "retrieving last request for all the model endpoints from the KV table.",
                    limit=V3IO_CLIENT_MEPS_LIMIT,
                    amount=len(endpoint_ids),
                )

                res = self.v3io_client.kv.new_cursor(
                    container=self.container,
                    table_path=self.last_request_table,
                ).all()
                last_request_timestamps.update(
                    {d["__name"]: d["last_request_timestamp"] for d in res}
                )
            else:
                filter_expression = " OR ".join(
                    [f"__name=='{endpoint_id}'" for endpoint_id in endpoint_ids]
                )
                res = self.v3io_client.kv.new_cursor(
                    container=self.container,
                    table_path=self.last_request_table,
                    filter_expression=filter_expression,
                ).all()
                last_request_timestamps.update(
                    {d["__name"]: d["last_request_timestamp"] for d in res}
                )
        except Exception as e:
            logger.warning(
                "Failed to get last request timestamp from V3IO KV table.",
                err=mlrun.errors.err_to_str(e),
                project=self.project,
                table=self.last_request_table,
            )

        return last_request_timestamps

    def get_drift_status(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        get_raw: bool = False,
    ) -> Union[pd.DataFrame, list[v3io_frames.client.RawFrame]]:
        filter_query = self._generate_filter_query(
            filter_key=mm_schemas.ApplicationEvent.ENDPOINT_ID,
            filter_values=endpoint_ids,
        )

        start, end = get_start_end(start, end, delta=timedelta(hours=24))
        res = self._get_records(
            table=mm_schemas.V3IOTSDBTables.APP_RESULTS,
            start=start,
            end=end,
            columns=[mm_schemas.ResultData.RESULT_STATUS],
            filter_query=filter_query,
            agg_funcs=["max"],
            group_by="endpoint_id",
            get_raw=get_raw,
        )
        if get_raw:
            return res

        df = res
        if not df.empty:
            df.columns = [
                col[len("max(") : -1] if "max(" in col else col for col in df.columns
            ]
        return df.reset_index(drop=True)

    def get_metrics_metadata(
        self,
        endpoint_id: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> pd.DataFrame:
        start, end = get_start_end(start, end)
        filter_query = self._generate_filter_query(
            filter_key=mm_schemas.ApplicationEvent.ENDPOINT_ID,
            filter_values=endpoint_id,
        )
        df = self._get_records(
            table=mm_schemas.V3IOTSDBTables.METRICS,
            start=start,
            end=end,
            columns=[mm_schemas.MetricData.METRIC_VALUE],
            filter_query=filter_query,
            agg_funcs=["last"],
        )
        if not df.empty:
            df.drop(
                columns=[f"last({mm_schemas.MetricData.METRIC_VALUE})"], inplace=True
            )
        return df.reset_index(drop=True)

    def get_results_metadata(
        self,
        endpoint_id: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> pd.DataFrame:
        start, end = get_start_end(start, end)
        filter_query = self._generate_filter_query(
            filter_key=mm_schemas.ApplicationEvent.ENDPOINT_ID,
            filter_values=endpoint_id,
        )
        df = self._get_records(
            table=mm_schemas.V3IOTSDBTables.APP_RESULTS,
            start=start,
            end=end,
            columns=[
                mm_schemas.ResultData.RESULT_KIND,
            ],
            filter_query=filter_query,
            agg_funcs=["last"],
        )
        if not df.empty:
            df.rename(
                columns={
                    f"last({mm_schemas.ResultData.RESULT_KIND})": mm_schemas.ResultData.RESULT_KIND
                },
                inplace=True,
            )
        return df.reset_index(drop=True)

    def get_error_count(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        get_raw: bool = False,
    ) -> Union[pd.DataFrame, list[v3io_frames.client.RawFrame]]:
        filter_query = self._generate_filter_query(
            filter_key=mm_schemas.ApplicationEvent.ENDPOINT_ID,
            filter_values=endpoint_ids,
        )
        if filter_query:
            filter_query += f"AND {mm_schemas.EventFieldType.ERROR_TYPE} == '{mm_schemas.EventFieldType.INFER_ERROR}'"
        else:
            filter_query = f"{mm_schemas.EventFieldType.ERROR_TYPE} == '{mm_schemas.EventFieldType.INFER_ERROR}' z"
        start, end = get_start_end(start, end)
        res = self._get_records(
            table=mm_schemas.FileTargetKind.ERRORS,
            start=start,
            end=end,
            columns=[mm_schemas.EventFieldType.ERROR_COUNT],
            filter_query=filter_query,
            agg_funcs=["count"],
            get_raw=get_raw,
        )

        if get_raw:
            return res

        df = res
        if not df.empty:
            df.rename(
                columns={
                    f"count({mm_schemas.EventFieldType.ERROR_COUNT})": mm_schemas.EventFieldType.ERROR_COUNT
                },
                inplace=True,
            )
            df.dropna(inplace=True)
        return df.reset_index(drop=True)

    def get_avg_latency(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        get_raw: bool = False,
    ) -> Union[pd.DataFrame, list[v3io_frames.client.RawFrame]]:
        filter_query = self._generate_filter_query(
            filter_key=mm_schemas.ApplicationEvent.ENDPOINT_ID,
            filter_values=endpoint_ids,
        )
        start = start or (mlrun.utils.datetime_now() - timedelta(hours=24))
        start, end = get_start_end(start, end)
        res = self._get_records(
            table=mm_schemas.V3IOTSDBTables.PREDICTIONS,
            start=start,
            end=end,
            columns=[mm_schemas.EventFieldType.LATENCY],
            filter_query=filter_query,
            agg_funcs=["avg"],
            get_raw=get_raw,
        )

        if get_raw:
            return res

        df = res
        if not df.empty:
            df.dropna(inplace=True)
            df.rename(
                columns={
                    f"avg({mm_schemas.EventFieldType.LATENCY})": f"avg_{mm_schemas.EventFieldType.LATENCY}"
                },
                inplace=True,
            )
        return df.reset_index(drop=True)

    def add_basic_metrics(
        self,
        model_endpoint_objects: list[mlrun.common.schemas.ModelEndpoint],
        metric_list: Optional[list[str]] = None,
    ) -> list[mlrun.common.schemas.ModelEndpoint]:
        """
        Fetch basic metrics from V3IO TSDB and add them to MEP objects.

        :param model_endpoint_objects: A list of `ModelEndpoint` objects that will
                                       be filled with the relevant basic metrics.
        :param metric_list:            List of metrics to include from the time series DB. Defaults to all metrics.

        :return: A list of `ModelEndpointMonitoringMetric` objects.
        """

        uids = []
        model_endpoint_objects_by_uid = {}
        for model_endpoint_object in model_endpoint_objects:
            uid = model_endpoint_object.metadata.uid
            uids.append(uid)
            model_endpoint_objects_by_uid[uid] = model_endpoint_object

        metric_name_to_function_and_column_name = {
            "error_count": (self.get_error_count, "count(error_count)"),
            "avg_latency": (self.get_avg_latency, "avg(latency)"),
            "result_status": (self.get_drift_status, "max(result_status)"),
        }
        if metric_list is not None:
            for metric_name in list(metric_name_to_function_and_column_name):
                if metric_name not in metric_list:
                    del metric_name_to_function_and_column_name[metric_name]

        metric_name_to_result = {}

        for metric_name, (
            function,
            _,
        ) in metric_name_to_function_and_column_name.items():
            metric_name_to_result[metric_name] = function(
                endpoint_ids=uids,
                get_raw=True,
            )

        def add_metric(
            metric: str,
            column_name: str,
            frames: list,
        ):
            for frame in frames:
                endpoint_ids = frame.column_data("endpoint_id")
                metric_data = frame.column_data(column_name)
                for index, endpoint_id in enumerate(endpoint_ids):
                    mep = model_endpoint_objects_by_uid.get(endpoint_id)
                    value = metric_data[index]
                    if mep and value is not None and not math.isnan(value):
                        setattr(mep.status, metric, value)

        for metric_name, result in metric_name_to_result.items():
            add_metric(
                metric_name,
                metric_name_to_function_and_column_name[metric_name][1],
                result,
            )
        if metric_list is None or "last_request" in metric_list:
            self._enrich_mep_with_last_request(
                model_endpoint_objects_by_uid=model_endpoint_objects_by_uid
            )

        return list(model_endpoint_objects_by_uid.values())

    def _enrich_mep_with_last_request(
        self,
        model_endpoint_objects_by_uid: dict[str, mlrun.common.schemas.ModelEndpoint],
    ):
        last_request_dictionary = self.get_last_request(
            endpoint_ids=list(model_endpoint_objects_by_uid.keys())
        )
        for uid, mep in model_endpoint_objects_by_uid.items():
            # Set the last request timestamp to the MEP object. If not found, keep the existing value from the
            # DB (relevant for batch EP).
            mep.status.last_request = last_request_dictionary.get(
                uid, mep.status.last_request
            )

    def count_results_by_status(
        self,
        start: Optional[Union[datetime, str]] = None,
        end: Optional[Union[datetime, str]] = None,
        endpoint_ids: Optional[Union[str, list[str]]] = None,
        application_names: Optional[Union[str, list[str]]] = None,
        result_status_list: Optional[list[int]] = None,
    ) -> dict[tuple[str, int], int]:
        start, end = get_start_end(start=start, end=end, delta=timedelta(hours=24))
        filter_query = ""
        if endpoint_ids:
            filter_query = self._generate_filter_query(
                filter_key=mm_schemas.ApplicationEvent.ENDPOINT_ID,
                filter_values=endpoint_ids,
            )
        if application_names:
            app_filter_query = self._generate_filter_query(
                filter_key=mm_schemas.ApplicationEvent.APPLICATION_NAME,
                filter_values=application_names,
            )
            if filter_query:
                filter_query += f" AND {app_filter_query}"
            else:
                filter_query = app_filter_query

        raw_frames: list[v3io_frames.client.RawFrame] = self._get_records(
            table=mm_schemas.V3IOTSDBTables.APP_RESULTS,
            start=start,
            end=end,
            columns=[
                mm_schemas.ResultData.RESULT_VALUE,
                mm_schemas.ResultData.RESULT_STATUS,
            ],
            filter_query=filter_query,
            get_raw=True,
        )

        if not raw_frames:
            return {}

        # Count occurrences by (application_name, result_status) from RawFrame objects
        count_dict = {}

        for frame in raw_frames:
            # Extract column data from each RawFrame
            app_name = frame.column_data(mm_schemas.ApplicationEvent.APPLICATION_NAME)[
                0
            ]
            statuses = frame.column_data(mm_schemas.ResultData.RESULT_STATUS)

            for status in statuses:
                # Filter by result status if specified
                if result_status_list and status not in result_status_list:
                    continue

                # Convert application name to lower case
                key = (app_name.lower(), status)

                # Update the count in the dictionary
                count_dict[key] = count_dict.get(key, 0) + 1
        return count_dict

    def count_processed_model_endpoints(
        self,
        start: Optional[Union[datetime, str]] = None,
        end: Optional[Union[datetime, str]] = None,
        application_names: Optional[Union[str, list[str]]] = None,
    ) -> dict[str, int]:
        start, end = get_start_end(start=start, end=end, delta=timedelta(hours=24))
        group_by_columns = [
            mm_schemas.ApplicationEvent.APPLICATION_NAME,
            mm_schemas.ApplicationEvent.ENDPOINT_ID,
        ]

        def get_application_endpoints_records(
            record_type: Literal["metrics", "results"],
        ):
            if record_type == "results":
                table_path = self.tables[mm_schemas.V3IOTSDBTables.APP_RESULTS]
            else:
                table_path = self.tables[mm_schemas.V3IOTSDBTables.METRICS]
            sql_query = self._get_sql_query(
                table_path=table_path,
                columns=[mm_schemas.WriterEvent.START_INFER_TIME],
                group_by_columns=group_by_columns,
                application_names=application_names,
            )
            return self.frames_client.read(
                backend=_TSDB_BE,
                start=start,
                end=end,
                query=sql_query,
            )

        df_results = get_application_endpoints_records("results")
        df_metrics = get_application_endpoints_records("metrics")

        if df_results.empty and df_metrics.empty:
            return {}

        # Combine the two dataframes and count unique endpoints per application
        combined_df = pd.concat([df_results, df_metrics], ignore_index=True)
        if combined_df.empty:
            return {}
        combined_df.drop_duplicates(subset=group_by_columns, inplace=True)

        grouped_df = combined_df.groupby(
            mm_schemas.WriterEvent.APPLICATION_NAME
        ).count()

        return grouped_df[mm_schemas.WriterEvent.ENDPOINT_ID].to_dict()

    def calculate_latest_metrics(
        self,
        start: Optional[Union[datetime, str]] = None,
        end: Optional[Union[datetime, str]] = None,
        application_names: Optional[Union[str, list[str]]] = None,
    ) -> list[
        Union[mm_schemas.ApplicationResultRecord, mm_schemas.ApplicationMetricRecord]
    ]:
        metric_list = []
        start, end = get_start_end(start=start, end=end, delta=timedelta(hours=24))

        # Get the latest results
        def get_latest_metrics_records(
            record_type: Literal["metrics", "results"],
        ) -> pd.DataFrame:
            group_by_columns = [mm_schemas.ApplicationEvent.APPLICATION_NAME]
            if record_type == "results":
                table_path = self.tables[mm_schemas.V3IOTSDBTables.APP_RESULTS]
                columns = [
                    f"last({mm_schemas.ResultData.RESULT_STATUS})",
                    f"last({mm_schemas.ResultData.RESULT_VALUE})",
                    f"last({mm_schemas.ResultData.RESULT_KIND})",
                ]
                group_by_columns += [
                    mm_schemas.ResultData.RESULT_NAME,
                ]
            else:
                table_path = self.tables[mm_schemas.V3IOTSDBTables.METRICS]
                columns = [f"last({mm_schemas.MetricData.METRIC_VALUE})"]
                group_by_columns += [
                    mm_schemas.MetricData.METRIC_NAME,
                ]
            sql_query = self._get_sql_query(
                table_path=table_path,
                columns=columns,
                group_by_columns=group_by_columns,
                application_names=application_names,
            )

            return self.frames_client.read(
                backend=_TSDB_BE,
                start=start,
                end=end,
                query=sql_query,
            )

        df_results = get_latest_metrics_records("results")
        df_metrics = get_latest_metrics_records("metrics")

        if df_results.empty and df_metrics.empty:
            return metric_list

        # Convert the results DataFrame to a list of ApplicationResultRecord
        def build_metric_objects() -> (
            list[
                Union[
                    mm_schemas.ApplicationResultRecord,
                    mm_schemas.ApplicationMetricRecord,
                ]
            ]
        ):
            metric_objects = []
            if not df_results.empty:
                df_results.rename(
                    columns={
                        f"last({mm_schemas.ResultData.RESULT_VALUE})": mm_schemas.ResultData.RESULT_VALUE,
                        f"last({mm_schemas.ResultData.RESULT_STATUS})": mm_schemas.ResultData.RESULT_STATUS,
                        f"last({mm_schemas.ResultData.RESULT_KIND})": mm_schemas.ResultData.RESULT_KIND,
                    },
                    inplace=True,
                )
                for _, row in df_results.iterrows():
                    metric_objects.append(
                        mm_schemas.ApplicationResultRecord(
                            result_name=row[mm_schemas.ResultData.RESULT_NAME],
                            kind=row[mm_schemas.ResultData.RESULT_KIND],
                            status=row[mm_schemas.ResultData.RESULT_STATUS],
                            value=row[mm_schemas.ResultData.RESULT_VALUE],
                        )
                    )
            if not df_metrics.empty:
                df_metrics.rename(
                    columns={
                        f"last({mm_schemas.MetricData.METRIC_VALUE})": mm_schemas.MetricData.METRIC_VALUE,
                    },
                    inplace=True,
                )

                for _, row in df_metrics.iterrows():
                    metric_objects.append(
                        mm_schemas.ApplicationMetricRecord(
                            metric_name=row[mm_schemas.MetricData.METRIC_NAME],
                            value=row[mm_schemas.MetricData.METRIC_VALUE],
                        )
                    )
            return metric_objects

        return build_metric_objects()

    def get_drift_data(
        self,
        start: datetime,
        end: datetime,
    ) -> mm_schemas.ModelEndpointDriftValues:
        table = mm_schemas.V3IOTSDBTables.APP_RESULTS
        start, end, interval = self._prepare_aligned_start_end(start, end)
        raw_frames: list[v3io_frames.client.RawFrame] = self._get_records(
            table=table,
            start=start,
            end=end,
            columns=[mm_schemas.ResultData.RESULT_STATUS],
            get_raw=True,
        )

        if not raw_frames:
            return mm_schemas.ModelEndpointDriftValues(values=[])

        aggregated_data = self._aggregate_raw_drift_data(
            raw_frames=raw_frames, start=start, end=end, interval=interval
        )
        if not aggregated_data:
            return mm_schemas.ModelEndpointDriftValues(values=[])

        # Filter to only include entries with max result_status >= 1
        filtered_data = [
            (endpoint_id, timestamp, max_status)
            for endpoint_id, timestamp, max_status in aggregated_data
            if max_status >= 1
        ]

        if not filtered_data:
            return mm_schemas.ModelEndpointDriftValues(values=[])

        return self._convert_drift_data_to_values(aggregated_data=filtered_data)

    @staticmethod
    def _aggregate_raw_drift_data(
        raw_frames: list[v3io_frames.client.RawFrame],
        start: datetime,
        end: datetime,
        interval: str,
    ) -> list[tuple[str, datetime, float]]:
        """
        Aggregate raw drift data from RawFrame objects.

        :param raw_frames: List of RawFrame objects containing drift data.
        :param start:      Start datetime for filtering data.
        :param end:        End datetime for filtering data.
        :param interval:   Time interval string (e.g., '5min') for aggregation

        :returns: list of tuples: (endpoint_id, timestamp, max_result_status)
        """
        if not raw_frames:
            return []

        # Parse interval to get timedelta
        interval_td = pd.Timedelta(interval)

        # Collect all data points from RawFrame objects
        data_points = []
        for frame in raw_frames:
            endpoint_id = frame.column_data(EventFieldType.ENDPOINT_ID)[0]
            result_statuses = frame.column_data(mm_schemas.ResultData.RESULT_STATUS)
            timestamps = frame.indices()[0].times

            # Combine data from this frame
            for i, (status, timestamp) in enumerate(zip(result_statuses, timestamps)):
                # V3IO TSDB returns timestamps in nanoseconds
                timestamp_dt = pd.Timestamp(
                    timestamp, unit="ns", tzinfo=timezone.utc
                ).to_pydatetime()

                # Filter by time window
                if start <= timestamp_dt < end:
                    data_points.append((endpoint_id, timestamp_dt, status))

        if not data_points:
            return []

        # Group by endpoint_id and time intervals, then find max status
        # Create time buckets aligned to start
        grouped_data = {}
        for endpoint_id, timestamp, status in data_points:
            # Calculate which interval bucket this timestamp falls into
            time_diff = timestamp - start
            bucket_index = int(time_diff / interval_td)
            bucket_start = start + (bucket_index * interval_td)

            key = (endpoint_id, bucket_start)
            if key not in grouped_data:
                grouped_data[key] = status
            else:
                # Keep the maximum status value
                grouped_data[key] = max(grouped_data[key], status)

        # Convert to list of tuples
        result = [
            (endpoint_id, timestamp, max_status)
            for (endpoint_id, timestamp), max_status in grouped_data.items()
        ]

        return result

    @staticmethod
    def _convert_drift_data_to_values(
        aggregated_data: list[tuple[str, datetime, float]],
    ) -> mm_schemas.ModelEndpointDriftValues:
        """
        Convert aggregated drift data to ModelEndpointDriftValues format.

        :param aggregated_data: List of tuples (endpoint_id, timestamp, max_result_status)
        :return: ModelEndpointDriftValues with counts of suspected and detected per timestamp
        """
        suspected_val = mm_schemas.constants.ResultStatusApp.potential_detection.value
        detected_val = mm_schemas.constants.ResultStatusApp.detected.value

        # Group by timestamp and result status, then count occurrences
        timestamp_status_counts = {}
        for _, timestamp, max_status in aggregated_data:
            key = (timestamp, max_status)
            timestamp_status_counts[key] = timestamp_status_counts.get(key, 0) + 1

        # Organize by timestamp with counts for suspected and detected
        timestamp_counts = {}
        for (timestamp, status), count in timestamp_status_counts.items():
            if timestamp not in timestamp_counts:
                timestamp_counts[timestamp] = {
                    "count_suspected": 0,
                    "count_detected": 0,
                }

            if status == suspected_val:
                timestamp_counts[timestamp]["count_suspected"] = count
            elif status == detected_val:
                timestamp_counts[timestamp]["count_detected"] = count

        # Convert to the expected format: list of (timestamp, count_suspected, count_detected)
        values = [
            (timestamp, counts["count_suspected"], counts["count_detected"])
            for timestamp, counts in sorted(timestamp_counts.items())
        ]

        return mm_schemas.ModelEndpointDriftValues(values=values)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import inspect
import socket
import time
from os import environ
from typing import Callable, Optional, Union

import mlrun.common.schemas
import mlrun.errors
import mlrun.k8s_utils
import mlrun.utils
import mlrun.utils.regex
from mlrun.errors import err_to_str

from ..config import config
from ..execution import MLClientCtx
from ..model import RunObject
from ..render import ipython_display
from ..utils import logger
from .base import FunctionStatus
from .kubejob import KubejobRuntime
from .local import exec_from_params, load_module
from .pod import KubeResourceSpec
from .utils import RunError, get_func_selector, log_std


class DaskSpec(KubeResourceSpec):
    _dict_fields = KubeResourceSpec._dict_fields + [
        "extra_pip",
        "remote",
        "service_type",
        "nthreads",
        "kfp_image",
        "node_port",
        "min_replicas",
        "max_replicas",
        "scheduler_timeout",
        "scheduler_resources",
        "worker_resources",
    ]

    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        volumes=None,
        volume_mounts=None,
        env=None,
        resources=None,
        build=None,
        default_handler=None,
        entry_points=None,
        description=None,
        replicas=None,
        image_pull_policy=None,
        service_account=None,
        image_pull_secret=None,
        extra_pip=None,
        remote=None,
        service_type=None,
        nthreads=None,
        kfp_image=None,
        node_port=None,
        min_replicas=None,
        max_replicas=None,
        scheduler_timeout=None,
        node_name=None,
        node_selector=None,
        affinity=None,
        scheduler_resources=None,
        worker_resources=None,
        priority_class_name=None,
        disable_auto_mount=False,
        pythonpath=None,
        workdir=None,
        tolerations=None,
        preemption_mode=None,
        security_context=None,
        state_thresholds=None,
        serving_spec=None,
        graph=None,
        parameters=None,
        track_models=None,
    ):
        super().__init__(
            command=command,
            args=args,
            image=image,
            mode=mode,
            volumes=volumes,
            volume_mounts=volume_mounts,
            env=env,
            resources=resources,
            replicas=replicas,
            image_pull_policy=image_pull_policy,
            service_account=service_account,
            build=build,
            default_handler=default_handler,
            entry_points=entry_points,
            description=description,
            image_pull_secret=image_pull_secret,
            node_name=node_name,
            node_selector=node_selector,
            affinity=affinity,
            priority_class_name=priority_class_name,
            disable_auto_mount=disable_auto_mount,
            pythonpath=pythonpath,
            workdir=workdir,
            tolerations=tolerations,
            preemption_mode=preemption_mode,
            security_context=security_context,
            state_thresholds=state_thresholds,
            serving_spec=serving_spec,
            graph=graph,
            parameters=parameters,
            track_models=track_models,
        )
        self.args = args

        self.extra_pip = extra_pip
        self.remote = True if remote is None else remote  # make remote the default

        self.service_type = service_type
        self.kfp_image = kfp_image
        self.node_port = node_port
        self.min_replicas = min_replicas or 0
        self.max_replicas = max_replicas or 16
        # supported format according to https://github.com/dask/dask/blob/master/dask/utils.py#L1402
        self.scheduler_timeout = scheduler_timeout or "60 minutes"
        self.nthreads = nthreads or 1
        self._scheduler_resources = self.enrich_resources_with_default_pod_resources(
            "scheduler_resources", scheduler_resources
        )
        self._worker_resources = self.enrich_resources_with_default_pod_resources(
            "worker_resources", worker_resources
        )

        self.state_thresholds = None  # not supported in dask

    @property
    def scheduler_resources(self) -> dict:
        return self._scheduler_resources

    @scheduler_resources.setter
    def scheduler_resources(self, resources):
        self._scheduler_resources = self.enrich_resources_with_default_pod_resources(
            "scheduler_resources", resources
        )

    @property
    def worker_resources(self) -> dict:
        return self._worker_resources

    @worker_resources.setter
    def worker_resources(self, resources):
        self._worker_resources = self.enrich_resources_with_default_pod_resources(
            "worker_resources", resources
        )


class DaskStatus(FunctionStatus):
    def __init__(
        self,
        state=None,
        build_pod=None,
        scheduler_address=None,
        cluster_name=None,
        node_ports=None,
    ):
        super().__init__(state, build_pod)

        self.scheduler_address = scheduler_address
        self.cluster_name = cluster_name
        self.node_ports = node_ports


class DaskCluster(KubejobRuntime):
    kind = "dask"
    _is_nested = False
    _is_remote = False

    def __init__(self, spec=None, metadata=None):
        super().__init__(spec, metadata)
        self._cluster = None
        self.use_remote = not mlrun.k8s_utils.is_running_inside_kubernetes_cluster()
        self.spec.build.base_image = (
            self.spec.build.base_image or mlrun.mlconf.default_base_image
        )

    @property
    def spec(self) -> DaskSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", DaskSpec)

    @property
    def status(self) -> DaskStatus:
        return self._status

    @status.setter
    def status(self, status):
        self._status = self._verify_dict(status, "status", DaskStatus)

    def is_deployed(self):
        if not self.spec.remote:
            return True
        return super().is_deployed()

    @property
    def initialized(self):
        return bool(self._cluster)

    def _load_db_status(self):
        meta = self.metadata
        if self._is_remote_api():
            db = self._get_db()
            db_func = None
            try:
                db_func = db.get_function(meta.name, meta.project, meta.tag)
            except Exception:
                pass

            if db_func and "status" in db_func:
                self.status = db_func["status"]
                if self.kfp:
                    logger.info(f"Dask status: {db_func['status']}")
                return "scheduler_address" in db_func["status"]

        return False

    def _start(self, watch=True):
        db = self._get_db()
        if not self._is_remote_api():
            self._cluster = db.start_function(function=self)
            return

        self.try_auto_mount_based_on_config()
        self._fill_credentials()
        if not self.is_deployed():
            raise RunError(
                "Function image is not built/ready, use .deploy()"
                " method first, or set base dask image to mlrun/mlrun"
            )

        self.save(versioned=False)
        background_task = db.start_function(func_url=self._function_uri())
        if watch:
            now = datetime.datetime.utcnow()
            timeout = now + datetime.timedelta(
                seconds=int(
                    mlrun.mlconf.background_tasks.default_timeouts.runtimes.dask_cluster_start
                )
            )
            while now < timeout:
                background_task = db.get_project_background_task(
                    background_task.metadata.project, background_task.metadata.name
                )
                if (
                    background_task.status.state
                    in mlrun.common.schemas.BackgroundTaskState.terminal_states()
                ):
                    if (
                        background_task.status.state
                        == mlrun.common.schemas.BackgroundTaskState.failed
                    ):
                        raise mlrun.errors.MLRunRuntimeError(
                            "Failed bringing up dask cluster"
                        )
                    else:
                        function = db.get_function(
                            self.metadata.name,
                            self.metadata.project,
                            self.metadata.tag,
                        )
                        if function and function.get("status"):
                            self.status = function.get("status")
                        return
                time.sleep(5)
                now = datetime.datetime.utcnow()
            raise mlrun.errors.MLRunTimeoutError(
                "Timeout waiting for Dask cluster to start"
            )

    def close(self, running=True):
        from dask.distributed import default_client

        try:
            client = default_client()
            # shutdown the cluster first, then close the client
            client.shutdown()
            client.close()
        except ValueError:
            pass

    def get_status(self):
        meta = self.metadata
        selector = get_func_selector(meta.project, meta.name, meta.tag)
        db = self._get_db()
        return db.function_status(meta.project, meta.name, self.kind, selector)

    def cluster(self):
        return self._cluster

    def _remote_addresses(self):
        addr = self.status.scheduler_address
        dash = ""
        if config.remote_host:
            if self.spec.service_type == "NodePort" and self.use_remote:
                addr = f"{config.remote_host}:{self.status.node_ports.get('scheduler')}"

            if self.spec.service_type == "NodePort":
                dash = f"{config.remote_host}:{self.status.node_ports.get('dashboard')}"
            else:
                logger.info("To get a dashboard link, use NodePort service_type")

        return addr, dash

    @property
    def client(self):
        from dask.distributed import Client, default_client

        if self.spec.remote and not self.status.scheduler_address:
            if not self._load_db_status():
                self._start()

        if self.status.scheduler_address:
            addr, dash = self._remote_addresses()
            logger.info(f"Trying dask client at: {addr}")
            try:
                client = Client(addr)
            except OSError as exc:
                logger.warning(
                    f"Remote scheduler at {addr} not ready, will try to restart {err_to_str(exc)}"
                )

                status = self.get_status()
                if status != "running":
                    self._start()
                addr, dash = self._remote_addresses()
                client = Client(addr)

            logger.info(
                f"Using remote dask scheduler ({self.status.cluster_name}) at: {addr}"
            )
            if dash:
                ipython_display(
                    f'<a href="http://{dash}/status" target="_blank" >dashboard link: {dash}</a>',
                    alt_text=f"remote dashboard: {dash}",
                )

            return client
        try:
            return default_client()
        except ValueError:
            return Client()

    def deploy(
        self,
        watch=True,
        with_mlrun=None,
        skip_deployed=False,
        is_kfp=False,
        mlrun_version_specifier=None,
        builder_env: Optional[dict] = None,
        show_on_failure: bool = False,
        force_build: bool = False,
    ):
        """deploy function, build container with dependencies

        :param watch:                   wait for the deploy to complete (and print build logs)
        :param with_mlrun:              add the current mlrun package to the container build
        :param skip_deployed:           skip the build if we already have an image for the function
        :param is_kfp:                  deploy as part of a kfp pipeline
        :param mlrun_version_specifier: which mlrun package version to include (if not current)
        :param builder_env:             Kaniko builder pod env vars dict (for config/credentials)
                                        e.g. builder_env={"GIT_TOKEN": token}
        :param show_on_failure:         show logs only in case of build failure
        :param force_build:             force building the image, even when no changes were made

        :return:                        True if the function is ready (deployed)
        """
        return super().deploy(
            watch,
            with_mlrun,
            skip_deployed,
            is_kfp=is_kfp,
            mlrun_version_specifier=mlrun_version_specifier,
            builder_env=builder_env,
            show_on_failure=show_on_failure,
            force_build=force_build,
        )

    def with_limits(
        self,
        mem=None,
        cpu=None,
        gpus=None,
        gpu_type="nvidia.com/gpu",
        patch: bool = False,
    ):
        raise NotImplementedError(
            "Use with_scheduler_limits/with_worker_limits to set resource limits",
        )

    def with_scheduler_limits(
        self,
        mem: Optional[str] = None,
        cpu: Optional[str] = None,
        gpus: Optional[int] = None,
        gpu_type: str = "nvidia.com/gpu",
        patch: bool = False,
    ):
        """
        set scheduler pod resources limits
        by default it overrides the whole limits section, if you wish to patch specific resources use `patch=True`.
        """
        self.spec._verify_and_set_limits(
            "scheduler_resources", mem, cpu, gpus, gpu_type, patch=patch
        )

    def with_worker_limits(
        self,
        mem: Optional[str] = None,
        cpu: Optional[str] = None,
        gpus: Optional[int] = None,
        gpu_type: str = "nvidia.com/gpu",
        patch: bool = False,
    ):
        """
        set worker pod resources limits
        by default it overrides the whole limits section, if you wish to patch specific resources use `patch=True`.
        """
        self.spec._verify_and_set_limits(
            "worker_resources", mem, cpu, gpus, gpu_type, patch=patch
        )

    def with_requests(self, mem=None, cpu=None, patch: bool = False):
        raise NotImplementedError(
            "Use with_scheduler_requests/with_worker_requests to set resource requests",
        )

    def with_scheduler_requests(
        self, mem: Optional[str] = None, cpu: Optional[str] = None, patch: bool = False
    ):
        """
        set scheduler pod resources requests
        by default it overrides the whole requests section, if you wish to patch specific resources use `patch=True`.
        """
        self.spec._verify_and_set_requests("scheduler_resources", mem, cpu, patch=patch)

    def with_worker_requests(
        self, mem: Optional[str] = None, cpu: Optional[str] = None, patch: bool = False
    ):
        """
        set worker pod resources requests
        by default it overrides the whole requests section, if you wish to patch specific resources use `patch=True`.
        """
        self.spec._verify_and_set_requests("worker_resources", mem, cpu, patch=patch)

    def set_state_thresholds(
        self,
        state_thresholds: dict[str, str],
        patch: bool = True,
    ):
        raise NotImplementedError(
            "State thresholds is not supported for Dask runtime yet, use spec.scheduler_timeout instead.",
        )

    def run(
        self,
        runspec: Optional[
            Union["mlrun.run.RunTemplate", "mlrun.run.RunObject", dict]
        ] = None,
        handler: Optional[Union[str, Callable]] = None,
        name: Optional[str] = "",
        project: Optional[str] = "",
        params: Optional[dict] = None,
        inputs: Optional[dict[str, str]] = None,
        out_path: Optional[str] = "",
        workdir: Optional[str] = "",
        artifact_path: Optional[str] = "",
        watch: Optional[bool] = True,
        schedule: Optional[Union[str, mlrun.common.schemas.ScheduleCronTrigger]] = None,
        hyperparams: Optional[dict[str, list]] = None,
        hyper_param_options: Optional[mlrun.model.HyperParamOptions] = None,
        verbose: Optional[bool] = None,
        scrape_metrics: Optional[bool] = None,
        local: Optional[bool] = False,
        local_code_path: Optional[str] = None,
        auto_build: Optional[bool] = None,
        param_file_secrets: Optional[dict[str, str]] = None,
        notifications: Optional[list[mlrun.model.Notification]] = None,
        returns: Optional[list[Union[str, dict[str, str]]]] = None,
        state_thresholds: Optional[dict[str, int]] = None,
        reset_on_run: Optional[bool] = None,
        output_path: Optional[str] = "",
        retry: Optional[Union[mlrun.model.Retry, dict]] = None,
        **launcher_kwargs,
    ) -> RunObject:
        if state_thresholds:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "State thresholds is not supported for Dask runtime yet, use spec.scheduler_timeout instead."
            )
        return super().run(
            runspec=runspec,
            handler=handler,
            name=name,
            project=project,
            params=params,
            inputs=inputs,
            out_path=out_path,
            workdir=workdir,
            output_path=output_path,
            watch=watch,
            schedule=schedule,
            hyperparams=hyperparams,
            hyper_param_options=hyper_param_options,
            verbose=verbose,
            scrape_metrics=scrape_metrics,
            local=local,
            local_code_path=local_code_path,
            auto_build=auto_build,
            param_file_secrets=param_file_secrets,
            notifications=notifications,
            returns=returns,
            state_thresholds=state_thresholds,
            retry=retry,
            **launcher_kwargs,
        )

    def _run(self, runobj: RunObject, execution):
        handler = runobj.spec.handler
        self._force_handler(handler)

        # TODO: investigate if the following instructions could overwrite the environment on any MLRun API Pod
        # Such action could result on race conditions against other runtimes and MLRun itself
        extra_env = self._generate_runtime_env(runobj)
        environ.update(extra_env)

        context = MLClientCtx.from_dict(
            runobj.to_dict(),
            rundb=self.spec.rundb,
            autocommit=False,
            host=socket.gethostname(),
        )
        if not inspect.isfunction(handler):
            if not self.spec.command:
                raise ValueError(
                    "specified handler (string) without command "
                    "(py file path), specify command or use handler pointer"
                )
            # Do not embed the module in system as it is not persistent with the dask cluster
            handler = load_module(
                self.spec.command,
                handler,
                context=context,
                embed_in_sys=False,
            )
        client = self.client
        setattr(context, "dask_client", client)
        sout, serr = exec_from_params(handler, runobj, context)
        log_std(self._db_conn, runobj, sout, serr, skip=self.is_child, show=False)
        return context.to_dict()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# for this function to run in nuclio you need to:
# - set the python base image e.g.:
#     python:3.6-jessie
# - add mlrun package install to the Build commands:
#     pip install mlrun


import time

from mlrun import get_or_create_ctx


def handler(context, event):
    ctx = get_or_create_ctx("myfunc", event=event)
    p1 = ctx.get_param("p1", 1)
    p2 = ctx.get_param("p2", "a-string")

    context.logger.info(
        f"Run: {ctx.name} uid={ctx.uid}:{ctx.iteration} Params: p1={p1}, p2={p2}"
    )

    time.sleep(1)

    # log scalar values (KFP metrics)
    ctx.log_result("accuracy", p1 * 2)
    ctx.log_result("latency", p1 * 3)

    # log various types of artifacts (and set UI viewers)
    ctx.log_artifact("test", body=b"abc is 123", local_path="test.txt")
    ctx.log_artifact("test_html", body=b"<b> Some HTML <b>", format="html")

    context.logger.info("run complete!")
    return ctx.to_json()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import requests
import requests.adapters
import requests.utils
import urllib3.exceptions
import urllib3.util.retry

from ..config import config
from ..errors import err_to_str
from . import logger


class HTTPSessionWithRetry(requests.Session):
    """
    Extend requests.Session to add retry logic on both error statuses and certain exceptions.
    """

    # make sure to only add exceptions that are raised early in the request. For example, ConnectionError can be raised
    # during the handling of a request, and therefore should not be retried, as the request might not be idempotent.

    # use strings because some requests exceptions are encapsulated in other exceptions, and we want to catch them all.
    HTTP_RETRYABLE_EXCEPTION_STRINGS = [
        # "Connection reset by peer" is raised when the server closes the connection prematurely during TCP handshake.
        "Connection reset by peer",
        # "Connection aborted" and "Connection refused" happen when the server doesn't respond at all.
        "Connection aborted",
        "Connection refused",
    ]

    # most of the exceptions would not be encapsulated, we want to catch them all directly.
    # this allows us more flexibility when deciding which *exactly* exception we want to retry on.
    HTTP_RETRYABLE_EXCEPTIONS = [
        # "Connection reset by peer" is raised when the server closes the connection prematurely during TCP handshake.
        ConnectionResetError,
        # "Connection aborted" and "Connection refused" happen when the server doesn't respond at all.
        ConnectionRefusedError,
        ConnectionAbortedError,
        ConnectionError,
        # often happens when the server is overloaded and can't handle the load.
        requests.exceptions.ConnectionError,
        requests.exceptions.ConnectTimeout,
        requests.exceptions.ReadTimeout,
        urllib3.exceptions.ReadTimeoutError,
        # may occur when connection breaks during the request.
        requests.exceptions.ChunkedEncodingError,
        urllib3.exceptions.InvalidChunkLength,
    ]

    def __init__(
        self,
        max_retries=config.http_retry_defaults.max_retries,
        retry_backoff_factor=config.http_retry_defaults.backoff_factor,
        retry_on_exception=True,
        retry_on_status=True,
        retry_on_post=False,
        verbose=False,
    ):
        """
        Initialize a new HTTP session with retry logic.
        :param max_retries:             Maximum number of retries to attempt.
        :param retry_backoff_factor:    Wait interval retries in seconds.
        :param retry_on_exception:      Retry on the HTTP_RETRYABLE_EXCEPTIONS. defaults to True.
        :param retry_on_status:         Retry on error status codes. defaults to True.
        :param retry_on_post:           Retry on POST requests. defaults to False.
        :param verbose:                 Print debug messages.
        """
        super().__init__()

        self.max_retries = max_retries
        self.retry_backoff_factor = retry_backoff_factor
        self.retry_on_exception = retry_on_exception
        self.verbose = verbose
        self._logger = logger.get_child("http-client")
        self._retry_methods = self._resolve_retry_methods(retry_on_post)

        if retry_on_status:
            self._http_adapter = requests.adapters.HTTPAdapter(
                max_retries=urllib3.util.retry.Retry(
                    total=self.max_retries,
                    backoff_factor=self.retry_backoff_factor,
                    status_forcelist=config.http_retry_defaults.status_codes,
                    allowed_methods=self._retry_methods,
                    # we want to retry but not to raise since we do want that last response (to parse details on the
                    # error from response body) we'll handle raising ourselves
                    raise_on_status=False,
                ),
                pool_maxsize=int(config.httpdb.max_workers),
            )

            self.mount("http://", self._http_adapter)
            self.mount("https://", self._http_adapter)

    def request(self, method, url, **kwargs):
        retry_count = 0
        kwargs.setdefault("headers", {})
        kwargs["headers"]["User-Agent"] = (
            f"{requests.utils.default_user_agent()} mlrun/{config.version}"
        )
        while True:
            try:
                response = super().request(method, url, **kwargs)
                return response
            except Exception as exc:
                if not self._error_is_retryable(url, method, exc, retry_count):
                    raise exc

                self._logger.warning(
                    "Error during request handling, retrying",
                    exc=err_to_str(exc),
                    retry_count=retry_count,
                    url=url,
                    method=method,
                )
                if self.verbose:
                    self._log_exception(
                        "debug",
                        exc,
                        f"{method} {url} request failed on retryable exception, "
                        f"retrying in {self.retry_backoff_factor} seconds",
                        retry_count,
                    )
                retry_count += 1
                time.sleep(self.retry_backoff_factor)

    def _error_is_retryable(self, url, method, exc, retry_count):
        if not self.retry_on_exception:
            self._log_exception(
                "warning",
                exc,
                f"{method} {url} request failed, http retries disabled,"
                f" raising exception: {err_to_str(exc)}",
                retry_count,
            )
            return False

        # if the response is not retryable, stop retrying
        # this is done to prevent the retry logic from running on non-idempotent methods (such as POST).
        if not self._method_retryable(method):
            self._log_exception(
                "warning",
                exc,
                f"{method} {url} request failed, http retries disabled for {method} method.",
                retry_count,
            )
            return False

        if retry_count >= self.max_retries:
            self._log_exception(
                "warning",
                exc,
                f"{method} {url} request failed, max retries reached,"
                f" raising exception: {err_to_str(exc)}",
                retry_count,
            )
            return False

        def exception_is_retryable(exc, retryable_exceptions):
            def err_chain(err):
                while err:
                    yield err
                    err = err.__cause__

            return any(
                isinstance(err_in_chain, retryable_exc)
                for retryable_exc in retryable_exceptions
                for err_in_chain in err_chain(exc)
            )

        # only retryable exceptions
        exception_is_retryable = exception_is_retryable(
            exc, self.HTTP_RETRYABLE_EXCEPTIONS
        )

        if not exception_is_retryable:
            self._log_exception(
                "warning",
                exc,
                f"{method} {url} request failed on non-retryable exception,"
                f" raising exception: {err_to_str(exc)}",
                retry_count,
            )
            return False
        return True

    def _method_retryable(self, method: str):
        return method in self._retry_methods

    def _resolve_retry_methods(self, retry_on_post: bool = False) -> frozenset[str]:
        methods = urllib3.util.retry.Retry.DEFAULT_ALLOWED_METHODS
        methods = methods.union({"PATCH"})
        if retry_on_post:
            methods = methods.union({"POST"})
        return frozenset(methods)

    def _log_exception(self, level, exc, message, retry_count):
        getattr(self._logger, level)(
            message,
            exception_type=type(exc),
            exception_message=err_to_str(exc),
            retry_interval=self.retry_backoff_factor,
            retry_count=retry_count,
            max_retries=self.max_retries,
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import tabulate

import mlrun.common.schemas
import mlrun.lists
import mlrun.utils.helpers

from .base import NotificationBase


class ConsoleNotification(NotificationBase):
    """
    Client only notification for printing run status notifications in console
    """

    def push(
        self,
        message: str,
        severity: typing.Optional[
            typing.Union[mlrun.common.schemas.NotificationSeverity, str]
        ] = mlrun.common.schemas.NotificationSeverity.INFO,
        runs: typing.Optional[typing.Union[mlrun.lists.RunList, list]] = None,
        custom_html: typing.Optional[typing.Optional[str]] = None,
        alert: typing.Optional[mlrun.common.schemas.AlertConfig] = None,
        event_data: typing.Optional[mlrun.common.schemas.Event] = None,
    ):
        severity = self._resolve_severity(severity)
        print(f"[{severity}] {message}")

        if not runs:
            return

        if isinstance(runs, list):
            runs = mlrun.lists.RunList(runs)

        table = []
        for run in runs:
            state = run["status"].get("state", "")
            if state == "error":
                result = run["status"].get("error", "")
            else:
                result = mlrun.utils.helpers.dict_to_str(
                    run["status"].get("results", {})
                )

            table.append(
                [
                    state,
                    run["metadata"]["name"],
                    ".." + run["metadata"]["uid"][-6:],
                    result,
                ]
            )
        print(tabulate.tabulate(table, headers=["status", "name", "uid", "results"]))

    def _resolve_severity(
        self, severity: typing.Union[mlrun.common.schemas.NotificationSeverity, str]
    ):
        if isinstance(severity, mlrun.common.schemas.NotificationSeverity):
            return severity
        elif isinstance(severity, str) and (
            (severity_lowercase := severity.lower())
            in set(mlrun.common.schemas.NotificationSeverity)
        ):
            return mlrun.common.schemas.NotificationSeverity(severity_lowercase)

        return severity.lower()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Optional, Union

import mlrun.errors
from mlrun.artifacts import Artifact, DatasetArtifact

from ..plan import MLPlan, MLPlanStages
from ..utils import MLTypes, MLUtils


class DatasetPlan(MLPlan):
    """
    A dataset plan for creating a dataset artifact. The plan can generate the artifact with a specific dataset according
    to the provided dataset split purpose: train, validation and test.
    """

    class Purposes(Enum):
        """
        All the dataset split purposes.
        """

        OTHER = "other"
        TRAIN = "train"
        VALIDATION = "validation"
        TEST = "test"

    # Default name dictionary to map a purpose to a default dataset artifact name:
    _DEFAULT_NAMES = {
        Purposes.OTHER: "dataset",
        Purposes.TRAIN: "train_set",
        Purposes.VALIDATION: "validation_set",
        Purposes.TEST: "test_set",
    }

    def __init__(
        self,
        purpose: Union[Purposes, str] = Purposes.OTHER,
        name: Optional[str] = None,
        preview: Optional[int] = None,
        stats: bool = False,
        fmt: str = "parquet",
    ):
        """
        Initialize a new dataset plan.

        :param purpose: The purpose of this dataset, can be one of DatasetPlan.Purposes:
                        * "other" will manually log the given dataset.
                        * "train" will log the training set (the one who sent to 'fit').
                        * "validation" will log the validation set (the one who set in 'apply_mlrun' as 'x_test' and
                          'y_test'.
                        * "test" will log the test set (the one who sent to 'predict').
        :param name:    The name to store the dataset as. Will be defaulted according to the purpose to be "dataset"
                        for "any", "train_set" for "train" and "test_set" for "test".
        :param preview: Number of lines to store as preview in the artifact metadata.
        :param stats:   Calculate and store dataset stats in the artifact metadata. Will be replaced soon with dataset
                        artifacts.
        :param fmt:     Format to use for saving the dataset. Can be one of {"csv", "parquet", "pq", "tsdb", "kv"}.

        :raise MLRunInvalidArgumentError: If either one of the arguments is invalid.
        """
        # Validate format:
        if fmt is not None and fmt not in DatasetArtifact.SUPPORTED_FORMATS:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"The given format: '{fmt}' is not supported. Supporting only the following formats: "
                f"{', '.join(DatasetArtifact.SUPPORTED_FORMATS)}"
            )

        # Store the configurations:
        self._purpose = (
            self.Purposes(purpose)
            if not isinstance(purpose, self.Purposes)
            else purpose
        )
        self._name = name if name is not None else self._DEFAULT_NAMES[self._purpose]
        self._preview = preview
        self._stats = stats
        self._fmt = fmt
        self._plans = {}  # TODO: Implement DatasetPlansLibrary with dataset specific artifacts plans.

        # Continue initializing the plan:
        super().__init__(need_probabilities=False)

    def is_ready(self, stage: MLPlanStages, is_probabilities: bool) -> bool:
        """
        Check whether or not the plan is fit for production by the given stage.

        :param stage:            The stage to check if the plan is ready.
        :param is_probabilities: True if the 'y_pred' that will be sent to 'produce' is a prediction of probabilities
                                 (from 'predict_proba') and False if not.

        :return: True if the plan is producible and False otherwise.
        """
        # For training set:
        if self._purpose == self.Purposes.TRAIN:
            return stage == MLPlanStages.PRE_FIT

        # For validation and test sets:
        if (
            self._purpose == self.Purposes.VALIDATION
            or self._purpose == self.Purposes.TEST
        ):
            return stage == MLPlanStages.PRE_PREDICT

        # For other:
        return True

    def produce(
        self,
        x: MLTypes.DatasetType,
        y: MLTypes.DatasetType = None,
        target_columns_names: MLTypes.TargetColumnsNamesType = None,
        **kwargs,
    ) -> dict[str, Artifact]:
        """
        Produce the dataset artifact according to this plan.

        :param x:                    A collection of inputs to a model.
        :param y:                    A collection of ground truth labels corresponding to the inputs.
        :param target_columns_names: List of names or indices to give the columns of the ground truth labels.

        :return: The produced dataset artifact.

        :raise MLRunInvalidArgumentError: If no dataset parameters were passed.
        """
        # Merge x and y into a single dataset:
        dataset, y_columns = MLUtils.concatenate_x_y(
            x=x, y=y, target_columns_names=target_columns_names
        )

        # Create the dataset artifact:
        dataset_artifact = DatasetArtifact(
            key=self._name,
            df=dataset,
            preview=self._preview,
            format=self._fmt,
            stats=self._stats,
        )

        # Add the purpose as a label:
        if self._purpose != self.Purposes.OTHER:
            dataset_artifact.labels["Purpose"] = self._purpose.value

        # TODO: Add the target columns names as an additional artifact (save as a json for example)

        # Store it:
        self._artifacts[self._name] = dataset_artifact

        return self._artifacts

    def _cli_display(self):
        """
        How the plan's products would be presented on a cli kernel.
        """
        print(self._artifacts[self._name].df)

    def _gui_display(self):
        """
        How the plan's products would be presented on a graphic IPython kernel (like a Jupyter notebook).
        """
        from IPython.display import display

        display(self._artifacts[self._name].df)
